# -*- coding: utf-8 -*
"""Python classes for enabling shortcut views on the biomedical sheet

Note that these shortcuts rely on the following assumptions.  These
assumptions are quite strong but make sense in clinical context.  Also, they
enforce a sensible streamlining of the high-throughput biomedical assay
structure which is the case for the interesting studies anyway.

- each shortcut sheet only uses one data type (e.g., WES, WGS, or mass-spec
  data)
- in the case of rare disease studies, only the first active bio and test
  sample and NGS library need to be considered
- in the case of cancer studies, only the first active test sample and NGS
  library are of interesting
- both one and multiple test samples can be of study, e.g., in the case
  of considering multiple parts of a larger tumor or metastesis
- the Decorator pattern is used for giving shortcut-based access to the sample
  sheet members

This all is to facilitate the automatic processing in pipelines.  Of course,
more assays can be combined by loading the results in a downstream step.
These downstream steps can then combine the information in any fashion they
see fit and build larger, more complex systems out of the simpler building
blocks that this module is meant for.

Also note that none of the shortcut objects will reflect changes in the
underlying schema data structure.
"""

from .. import models

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

#: Token for identifying a germline variants sequencing experiment
SHEET_TYPE_GERMLINE_VARIANTS = 'germline_variants'
#: Token for identifying a cancer matched tumor/normal study
SHEET_TYPE_CANCER_MATCHED = 'cancer_matched'
#: Token for identifying a generic sample sheet
SHEET_TYPE_GENERIC = 'generic'

#: Extraction type "DNA"
EXTRACTION_TYPE_DNA = 'DNA'
#: Extraction type "RNA"
EXTRACTION_TYPE_RNA = 'RNA'


class MissingDataWarning(UserWarning):
    """Used for warning for missing data in non-fatal cases"""


def is_background(sheet):
    """Return ``True`` if the ``Sheet`` shortcut is not flagged as background
    """
    return sheet.extra_infos.get('is_background', False)


def is_not_background(sheet):
    """Return ``True`` if the ``Sheet`` shortcut is not flagged as background
    """
    return not is_background(sheet)


class InvalidSelector(Exception):
    """Raised in the case of an invalid ``TestSample`` child type"""


class MissingDataEntity(Exception):
    """Raised if the given data entity is not known"""


class ShortcutSampleSheet:
    """Shortcut of a generic experiment sample sheet"""

    def __init__(self, sheet):
        #: The wrapped ``Sheet``
        self.sheet = sheet
        #: Shortcut to wrapped sheet
        self.wrapped = sheet
        #: A list of shortcut objects to all enabled NGS libraries
        self.all_ngs_libraries = None
        #: A list of shortcut objects to the first enabled NGS library for
        #: each sample
        self.primary_ngs_libraries = None

    @property
    def extra_infos(self):
        """Shorcut to wrapped object's ``extra_infos``"""
        return self.wrapped.extra_infos


class ShortcutMixin:
    """Mixin with helper functions"""

    @property
    def pk(self):
        """Shortcut to ``pk`` property of wrapped object

        The value is usually generated by data management system/database.
        """
        return self.wrapped.pk

    @property
    def secondary_id(self):
        """Shortcut to ``secondary_id`` property of wrapped object

        The value is usually assigned by the data generator/customer
        """
        return self.wrapped.secondary_id

    @property
    def name(self):
        """Return name generated for wrapped object"""
        return self.wrapped.name

    @property
    def disabled(self):
        """Return whether the entry has been disabled"""
        return bool(self.wrapped.disabled)

    @property
    def enabled(self):
        """Return whether the entry has been enabled"""
        return not self.disabled

    @property
    def extra_infos(self):
        """Shorcut to wrapped object's ``extra_infos``"""
        return self.wrapped.extra_infos


class BioEntityShortcut:
    pass


class BioSampleShortcut(ShortcutMixin):
    """When navigating with the ``biomedsheets.shortcuts`` through the
    sheets, this is the type that you end up instead of raw ``TestSample``
    objects

    Objects of this class are pre-configured with a specific ``TestSample``
    child type (e.g., NGS library) where the primary active will be picked.
    """

    def __init__(self, bio_entity, bio_sample, selector):
        #: Containing shortcut to ``BioEntity``
        self.bio_entity = bio_entity
        #: Raw ``BioEntity``
        self.bio_sample = bio_sample
        #: Wrapped for Shortcut Mixin
        self.wrapped = bio_sample
        # Check selector for being valid
        if selector not in models.TEST_SAMPLE_CHILDREN:
            raise InvalidSelector(
                'Invalid test sample selector {}'.format(selector))
        #: Selector for ``TestSample`` children
        self.selector = selector
        #: The selected ``TestSample`` with the selected ``TestSample`` child
        self.test_sample = self._get_test_sample()
        #: The selected ``TestSample`` child
        self.assay_sample = self.test_sample.assay_sample

    def _get_test_sample(self):
        """Return appropriate ``TestSampleShortcut`` raise an exception"""
        values = {
            models.KEY_NGS_LIBRARY: lambda x: x.ngs_libraries.values(),
        }
        for test_sample in self.bio_sample.test_samples.values():
            attr_lst = values.get(self.selector, lambda x: [])(test_sample)
            for entity in attr_lst:
                if not entity.disabled:
                    return TestSampleShortcut(self, test_sample, self.selector)
        raise MissingDataEntity(
            'Could not find data entity for type {} in {}'.format(
                self.selector, self.bio_sample))

    def __repr__(self):
        return 'BioSampleShortcut({})'.format(', '.join(map(str, [
            self.bio_sample, self.selector, self.test_sample,
            self.assay_sample])))

    def __str__(self):
        return repr(self)


class TestSampleShortcut:
    """When navigating with the ``biomedsheets.shortcuts`` through the
    sheets, this is the type that you end up instead of raw ``TestSample``
    objects

    Objects of this class are pre-configured with a specific ``TestSample``
    child type (e.g., NGS library) where the primary active one will be picked.
    """

    def __init__(self, bio_sample, test_sample, selector):
        #: Containing shortcut to ``BioSample``
        self.bio_sample = bio_sample
        #: Raw ``TestSample``
        self.test_sample = test_sample
        # Check selector for being valid
        if selector not in models.TEST_SAMPLE_CHILDREN:
            raise InvalidSelector(
                'Invalid test sample selector {}'.format(selector))
        #: Selector for ``TestSample`` children
        self.selector = selector
        #: The selected ``TestSample`` child
        self.assay_sample = self._get_assay_sample()

    @property
    def pk(self):
        """Shortcut to ``pk`` property of wrapped ``TestSample``

        The value is usually generated by data management system/database.
        """
        return self.test_sample.pk

    @property
    def secondary_id(self):
        """Shortcut to ``secondary_id`` property of wrapped ``TestSample``

        The value is usually assigned by the data generator/customer
        """
        return self.test_sample.secondary_id

    @property
    def name(self):
        """Return name generated for wrapped test sample"""
        return self.test_sample.name

    @property
    def disabled(self):
        """Shortcut to ``disabled`` property of wrapped ``TestSample``"""
        return self.test_sample.disabled

    @property
    def enabled(self):
        """Shortcut to ``enabled`` property of wrapped ``TestSample``"""
        return self.test_sample.enabled

    def _get_assay_sample(self):
        """Return ``TestSample`` child or raise an exception"""
        values = {
            models.KEY_NGS_LIBRARY: lambda x: x.ngs_libraries.values(),
        }
        constructors = {
            models.KEY_NGS_LIBRARY: NGSLibraryShortcut,
        }
        attr_lst = values.get(self.selector, lambda x: [])(self.test_sample)
        for entity in attr_lst:
            if not entity.disabled:
                return constructors[self.selector](self, entity)
        raise MissingDataEntity(
            'Could not find data entity for type {} in {}'.format(
                self.selector, self.test_sample))

    def __repr__(self):
        return 'TestSampleShortcut({})'.format(', '.join(map(str, [
            self.test_sample, self.selector, self.assay_sample])))

    def __str__(self):
        return repr(self)


class TestSampleChildShortcut:
    """Helper base class for children of ``TestSampleShortcut``
    """

    def __init__(self, test_sample, wrapped):
        #: Containing ``TestSampleChildShortcut``
        self.test_sample = test_sample
        #: Wrapped raw TestSample child
        self.wrapped = wrapped

    @property
    def pk(self):
        """Shortcut to ``pk`` property of wrapped ``TestSample``

        The value is usually generated by data management system/database.
        """
        return self.wrapped.pk

    @property
    def secondary_id(self):
        """Shortcut to ``secondary_id`` property of wrapped ``TestSample``

        The value is usually assigned by the data generator/customer
        """
        return self.wrapped.secondary_id

    @property
    def name(self):
        """Return name generated for wrapped object"""
        return self.wrapped.name

    @property
    def disabled(self):
        """Shortcut to ``disabled`` property of wrapped ``TestSample``"""
        return self.wrapped.disabled

    @property
    def enabled(self):
        """Shortcut to ``enabled`` property of wrapped ``TestSample``"""
        return self.wrapped.enabled

    @property
    def extra_infos(self):
        """Shorcut to wrapped object's ``extra_infos``"""
        return self.wrapped.extra_infos


class NGSLibraryShortcut(TestSampleChildShortcut):
    """Shortcut to NGSLibrary
    """

    def __init__(self, test_sample, ngs_library):
        super().__init__(test_sample, ngs_library)
        #: Wrapped raw ``NGSLibrary``
        self.ngs_library = ngs_library



# -*- coding: utf-8 -*-
"""Entrypoint for biomedsheets

The biomedsheets package provides Python library modules as well as an
executable for the various
"""

import argparse
import collections
import json
import logging
import os
import sys

import pkg_resources

from .io import SheetSchema, json_loads_ordered
from .io_tsv import (
    read_cancer_tsv_json_data, read_germline_tsv_json_data,
    read_generic_tsv_json_data)
from .ref_resolver import RefResolver
from .validation import SchemaValidator
from .shortcuts import (
    SHEET_TYPE_GERMLINE_VARIANTS, SHEET_TYPE_CANCER_MATCHED,
    SHEET_TYPE_GENERIC)

#: Choices for the TSV sheet type
CHOICES_SHEET_TYPE = (
    SHEET_TYPE_GERMLINE_VARIANTS, SHEET_TYPE_CANCER_MATCHED,
    SHEET_TYPE_GENERIC)

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


class AppBase:
    """Base class for application, sub classed for each command"""

    def __init__(self, args):
        #: Parsed command line arguments
        self.args = args


class JsonSheetAppBase(AppBase):
    """Base class for working with sheets"""

    def __init__(self, args):
        super().__init__(args)
        #: SheetSchema from bundled json
        self.sheet_schema = self.load_sheet_schema()
        #: Validator for BioMed JSON schemas
        self.validator = SchemaValidator(self.sheet_schema)
        #: Sample sheet JSON
        self.sheet_json = self.load_sheet_json()
        #: Resolved JSON
        self.resolved_json = self.resolve_refs(self.sheet_json)

    def run(self):
        """Execute the command"""
        raise NotImplementedError('Abstract method called: override me!')

    def load_sheet_schema(self):
        """Load the bundled SheetSchema"""
        print('Loading bundled BioMed Sheet JSON Schema...', file=sys.stderr)
        return SheetSchema.load_from_string(pkg_resources.resource_stream(
            'biomedsheets', 'data/sheet.schema.json').read().decode())

    def load_sheet_json(self):
        """Load the sheet JSON file"""
        print('Loading sheet JSON from "{}"'.format(self.args.input),
              file=sys.stderr)
        try:
            with open(self.args.input, 'rt') as f:
                return json_loads_ordered(f.read())
        except FileNotFoundError:
            raise  # re-raise

    def get_lib_dirs(self):
        """Return list of paths to search for relative paths"""
        lib_dirs = list(self.args.lib_dir)
        lib_dirs.append(os.path.dirname(os.path.abspath(self.args.input)))
        return lib_dirs

    def resolve_refs(self, sheet_json):
        """Resolve "$ref" JSON pointers in sheet_json"""
        print('Resolving {{ "$ref": "..." }} in JSON...')
        resolver = RefResolver(
            lookup_paths=self.get_lib_dirs(),
            dict_class=collections.OrderedDict)
        return resolver.resolve('file://' + self.args.input, sheet_json)

    def validate_and_print_errors(self):
        errors = self.validator.validate(self.resolved_json)
        if errors:
            print('The following validation errors occured', file=sys.stderr)
            for error in errors:
                print(' - {}'.format(error), file=sys.stderr)
        return errors


class ValidateApp(JsonSheetAppBase):
    """App sub class for validation sub command"""

    def run(self):
        if self.validate_and_print_errors():
            return 1
        else:
            print('Sheet passed validation, congratulation!', file=sys.stderr)


class ExpandApp(JsonSheetAppBase):
    """App sub class for expanding "$ref" JSON pointers"""

    def run(self):
        json.dump(self.resolved_json, self.args.output, indent='  ')
        print('', file=self.args.output)


class ConvertApp(AppBase):
    """App sub class for converting shortcut TSV sheets to JSON sheets"""

    def run(self):
        funcs = {
            SHEET_TYPE_GERMLINE_VARIANTS: read_germline_tsv_json_data,
            SHEET_TYPE_CANCER_MATCHED: read_cancer_tsv_json_data,
            SHEET_TYPE_GENERIC: read_generic_tsv_json_data,
        }
        json.dump(
            funcs[self.args.type](self.args.input, self.args.input.name),
            self.args.output,
            indent='    ')
        print('', file=self.args.output)


def run(args):
    """Program entry point after parsing command line arguments"""
    apps = {
        'validate': ValidateApp,
        'expand': ExpandApp,
        'convert': ConvertApp,
    }
    return apps[args.subparser](args).run()


def main(argv=None):
    """Main program entry point, starts parsing command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--lib-dir', type=str, default=['.', os.getcwd()], action='append',
        help=('Base directorie for JSON file pointers, can be given '
              'multiple times'))

    subparsers = parser.add_subparsers(dest='subparser')

    parser_validate = subparsers.add_parser(
        'validate', help='Validate sample sheet JSON')
    parser_validate.add_argument(
        '-i', '--input', type=str, required=True,
        help='Path to BioMed Sheet (JSON or YAML) file')

    parser_expand = subparsers.add_parser(
        'expand', help='Expand "$ref" JSON pointers')
    parser_expand.add_argument(
        '-i', '--input', type=str, required=True,
        help='Path to BioMed Sheet (JSON or YAML) file')
    parser_expand.add_argument(
        '-o', '--output', type=argparse.FileType('wt'), default=sys.stdout,
        help='Path to output file, defaults to stdout')

    parser_convert = subparsers.add_parser(
        'convert', help='Convert shortcut TSV sheet to JSON sample sheet')
    parser_convert.add_argument(
        '-t', '--type', choices=CHOICES_SHEET_TYPE, required=True,
        help='Shortcut TSV sheet type')
    parser_convert.add_argument(
        '-i', '--input', type=argparse.FileType('rt'), required=True,
        help='Path to input TSV file')
    parser_convert.add_argument(
        '-o', '--output', type=argparse.FileType('wt'), default=sys.stdout,
        help='Path to output file, defaults to stdout')

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)

    args = parser.parse_args(argv)
    if not args.subparser:
        parser.print_usage()
        return 1
    return run(args)


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
"""Code for reading and writing the compact generic TSV format.
"""

from collections import OrderedDict

from .base import (
    LIBRARY_TYPES, EXTRACTION_TYPES, std_field, TSVSheetException,
    BaseTSVReader)
from ..naming import name_generator_for_scheme, NAMING_DEFAULT

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

#: Default title
GENERIC_DEFAULT_TITLE = 'Generic Sample Sheet'

#: Default description
GENERIC_DEFAULT_DESCRIPTION = (
    'Sample Sheet constructed from generic compact TSV file')

#: Generic body TSV header
GENERIC_BODY_HEADER = (
    'bioEntity', 'bioSample', 'testSample', 'ngsLibrary', 'extractionType',
    'libraryType', 'folderName')

#: Fixed "extraInfoDefs" field for generic compact TSV
GENERIC_EXTRA_INFO_DEFS = OrderedDict([
    ('bioEntity', OrderedDict([
    ])),
    ('bioSample', OrderedDict([
    ])),
    ('testSample', OrderedDict([
        std_field('extractionType'),
    ])),
    ('ngsLibrary', OrderedDict([
        std_field('seqPlatform'),
        std_field('libraryType'),
        std_field('folderName'),
    ])),
])


class GenericTSVSheetException(TSVSheetException):
    """Raised on problems with loading generic TSV sample sheets"""


class GenericTSVReader(BaseTSVReader):
    """Helper class for reading generic TSV file

    Prefer using ``read_generic_tsv_*()`` for shortcut
    """

    body_header = GENERIC_BODY_HEADER
    extra_info_defs = GENERIC_EXTRA_INFO_DEFS
    default_title = GENERIC_DEFAULT_TITLE
    default_description = GENERIC_DEFAULT_DESCRIPTION
    bio_entity_name_column = 'bioEntity'
    bio_sample_name_column = 'bioSample'
    test_sample_name_column = 'testSample'
    ngs_library_name_column = 'ngsLibrary'

    def check_tsv_line(self, mapping, lineno):
        """Cancer sample sheet--specific valiation"""
        # Check for hyphen in patient or sample name
        for key in ('bioEntity', 'bioSample', 'testSample', 'ngsLibrary'):
            if '-' in mapping[key]:
                raise GenericTSVSheetException(  # pragma: no cover
                    'Hyphen not allowed in {} column'.format(key))
        # Check "extractionType" field
        if mapping['extractionType'] and (
                mapping['extractionType'] not in EXTRACTION_TYPES):
            raise GenericTSVSheetException(
                'Invalid extraction type {}, must be in {{{}}}'.format(
                    mapping['extractionType'], ', '.join(EXTRACTION_TYPES)))
        # Check "libraryType" field
        if mapping['libraryType'] and (
                mapping['libraryType'] not in LIBRARY_TYPES):
            raise GenericTSVSheetException(
                'Invalid library type {}, must be in {{{}}}'.format(
                    mapping['libraryType'], ', '.join(LIBRARY_TYPES)))

    @classmethod
    def _check_consistency(cls, records, key):
        values = list(sorted(set(r[key] for r in records)))
        if len(values) > 1:
            raise ValueError(  # pragma: no cover
                'Inconsistent {} entries in records: {}'.format(key, values))


def read_generic_tsv_sheet(f, fname=None, naming_scheme=NAMING_DEFAULT):
    """Read compact generic TSV format from file-like object ``f``

    :return: models.Sheet
    """
    return GenericTSVReader(f, fname).read_sheet(
        name_generator_for_scheme(naming_scheme))


def read_generic_tsv_json_data(f, fname=None):
    """Read compact generic TSV format from file-like object ``f``

    :return: ``dict``
    """
    return GenericTSVReader(f, fname).read_json_data()



# -*- coding: utf-8 -*-
"""Union-find data structure"""


class UnionFind:
    """Union-find data structure.

    Each unionFind instance X maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - X[item] returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in X, a new singleton set is
      created for it.

    - X.union(item1, item2, ...) merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in X, it is added to X as one of the members of the merged set.

    Adapted from https://www.ics.uci.edu/~eppstein/PADS/UnionFind.py

    Copyright (c) 2002-2015, David Eppstein
    """

    def __init__(self):
        """Create a new empty union-find structure"""
        self.weights = {}
        self.parents = {}

    def __getitem__(self, key):
        """Find and return the name of the set containing the key"""

        # check for previously unknown key
        if key not in self.parents:
            self.parents[key] = key
            self.weights[key] = 1
            return key

        # find path of keys leading to the root
        path = [key]
        root = self.parents[key]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def __iter__(self):
        """Iterate through all items ever found or unioned by this structure
        """
        return iter(self.parents)

    def union(self, *keys):
        """Find the sets containing the keys and merge them all"""
        roots = [self[x] for x in keys]
        heaviest = max([(self.weights[r], r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest



# -*- coding: utf-8 -*-
"""Python code for the validation of BioMed Schemas
"""

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


class SchemaValidator:
    """Main validation driver for JSON documents"""

    def __init__(self, sheet_schema):
        #: ``SheetShema`` to use for the validation
        self.sheet_schema = sheet_schema

    def validate(self, resolved_json):
        """Perform validation

        resolved_json JSON document in Python representation
        """
        pass



# -*- coding: utf-8 -*-
"""Tests for loading succing cancer TSV files with custom fields
"""

import io
import json
import textwrap

import pytest

from biomedsheets import io_tsv

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def tsv_sheet_cancer_header():
    f = io.StringIO(textwrap.dedent("""
    [Metadata]
    schema\tcancer_matched
    schema_version\tv1
    title\tExample matched cancer tumor/normal study
    description\tThe study has two patients, P001 has one tumor sample, P002 has two

    [Custom Fields]
    key\tannotatedEntity\tdocs\ttype\tminimum\tmaximum\tunit\tchoices\tpattern
    ncbiTaxon\tbioEntity\tNCBI taxon ID\tpattern\t.\t.\t.\t.\t^NCBITaxon_[1-9][0-9]*$
    mycnCn\tbioSample\tMYCN copy number\tnumber\t0.0\t.\t.\t.\t.
    inssStage\tbioSample\tInternational Neuroblastoma Staging System\tenum\t.\t.\t.\t1,2,2A,2B,3,4,4S\t.
    shimada4Class\tbioEntity\tShimada 4 classification\tenum\t.\t.\t.\tfavorable,unfavorable\t.
    lohChr1p\tbioSample\tChromosome arm 1p deleted\tboolean\t.\t.\t.\t.\t.
    patientStatus\tbioSample\tStatus of patient\tenum\t.\t.\t.\tDeceased_from_disease,No_disease_event,Relapse_progression\t.
    timeToEventDays\tbioEntity\tTime to event in days\tinteger\t0\t.\td\t.\t.
    timeToDeathDays\tbioEntity\tTime to death in days\tinteger\t0\t.\td\t.\t.
    testSampleNice\ttestSample\tIs test sample nice?\tboolean\t.\t.\t.\t.\t.
    ngsLibraryNice\tngsLibrary\tIs NGS library nice?\tboolean\t.\t.\t.\t.\t.

    [Data]
    patientName\tsampleName\tisTumor\tlibraryType\tfolderName\tncbiTaxon\tmycnCn\tinssStage\tshimada4Class\tlohChr1p\tpatientStatus\ttimeToEventDays\ttimeToDeathDays\ttestSampleNice\tngsLibraryNice
    P001\tN1\tN\tWES\tP001-N1-DNA1-WES1\tNCBITaxon_9606\t.\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\ttrue
    P001\tT1\tY\tWES\tP001-T1-DNA1-WES1\tNCBITaxon_9606\t2\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\tfalse
    P001\tT1\tY\tmRNA_seq\tP001-T1-RNA1-mRNAseq1\tNCBITaxon_9606\t2\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\ttrue
    P002\tN1\tN\tWES\tP001-N1-DNA1-WES1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\tfalse\ttrue
    P002\tT1\tY\tWES\tP001-T1-DNA1-WES1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    P002\tT1\tY\tWES\tP001-T1-RNA1-RNAseq1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\tfalse\tfalse
    P002\tT2\tY\tWES\tP001-T2-DNA1-WES1\tNCBITaxon_9606\t.\t3\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    P002\tT2\tY\tmRNA_seq\tP001-T2-RNA1-mRNAseq1\tNCBITaxon_9606\t.\t3\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    """.lstrip()))
    return f


@pytest.fixture
def tsv_sheet_cancer_no_header():
    """Tumor TSV sheet without header"""
    f = io.StringIO(textwrap.dedent("""
    patientName\tsampleName\tisTumor\tlibraryType\tfolderName\tncbiTaxon\tmycnCn\tinssStage\tshimada4Class\tlohChr1p\tpatientStatus\ttimeToEventDays\ttimeToDeathDays\ttestSampleNice\tngsLibraryNice
    P001\tN1\tN\tWES\tP001-N1-DNA1-WES1\tNCBITaxon_9606\t.\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\ttrue
    P001\tT1\tY\tWES\tP001-T1-DNA1-WES1\tNCBITaxon_9606\t2\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\tfalse
    P001\tT1\tY\tmRNA_seq\tP001-T1-RNA1-mRNAseq1\tNCBITaxon_9606\t2\t1\tfavorable\ttrue\tNo_disease_event\t.\t.\ttrue\ttrue
    P002\tN1\tN\tWES\tP001-N1-DNA1-WES1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\tfalse\ttrue
    P002\tT1\tY\tWES\tP001-T1-DNA1-WES1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    P002\tT1\tY\tWES\tP001-T1-RNA1-RNAseq1\tNCBITaxon_9606\t.\t2\tunfavorable\ttrue\tRelapse_progression\t.\t.\tfalse\tfalse
    P002\tT2\tY\tWES\tP001-T2-DNA1-WES1\tNCBITaxon_9606\t.\t3\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    P002\tT2\tY\tmRNA_seq\tP001-T2-RNA1-mRNAseq1\tNCBITaxon_9606\t.\t3\tunfavorable\ttrue\tRelapse_progression\t.\t.\ttrue\ttrue
    """.lstrip()))
    return f


# Expected value for the cancer sheet JSON with header
EXPECTED_CANCER_SHEET_JSON_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Example matched cancer tumor/normal study",
    "description": "The study has two patients, P001 has one tumor sample, P002 has two",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            },
            "shimada4Class": {
                "docs": "Shimada 4 classification",
                "key": "shimada4Class",
                "type": "enum",
                "choices": [
                    "favorable",
                    "unfavorable"
                ]
            },
            "timeToEventDays": {
                "docs": "Time to event in days",
                "key": "timeToEventDays",
                "type": "integer",
                "minimum": 0,
                "unit": "d"
            },
            "timeToDeathDays": {
                "docs": "Time to death in days",
                "key": "timeToDeathDays",
                "type": "integer",
                "minimum": 0,
                "unit": "d"
            }
        },
        "bioSample": {
            "isTumor": {
                "docs": "Boolean flag for distinguishing tumor/normal samples",
                "key": "isTumor",
                "type": "boolean"
            },
            "mycnCn": {
                "docs": "MYCN copy number",
                "key": "mycnCn",
                "type": "number",
                "minimum": 0.0
            },
            "inssStage": {
                "docs": "International Neuroblastoma Staging System",
                "key": "inssStage",
                "type": "enum",
                "choices": [
                    "1",
                    "2",
                    "2A",
                    "2B",
                    "3",
                    "4",
                    "4S"
                ]
            },
            "lohChr1p": {
                "docs": "Chromosome arm 1p deleted",
                "key": "lohChr1p",
                "type": "boolean"
            },
            "patientStatus": {
                "docs": "Status of patient",
                "key": "patientStatus",
                "type": "enum",
                "choices": [
                    "Deceased_from_disease",
                    "No_disease_event",
                    "Relapse_progression"
                ]
            }
        },
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            },
            "testSampleNice": {
                "docs": "Is test sample nice?",
                "key": "testSampleNice",
                "type": "boolean"
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            },
            "ngsLibraryNice": {
                "docs": "Is NGS library nice?",
                "key": "ngsLibraryNice",
                "type": "boolean"
            }
        }
    },
    "bioEntities": {
        "P001": {
            "pk": 1,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606",
                "shimada4Class": "favorable"
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {
                        "inssStage": "1",
                        "lohChr1p": true,
                        "patientStatus": "No_disease_event",
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 5,
                    "extraInfo": {
                        "isTumor": true,
                        "mycnCn": 2.0,
                        "inssStage": "1",
                        "lohChr1p": true,
                        "patientStatus": "No_disease_event"
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 6,
                            "extraInfo": {
                                "extractionType": "DNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 7,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": false
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 8,
                            "extraInfo": {
                                "extractionType": "RNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 9,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "P002": {
            "pk": 10,
            "extraInfo": {
                "ncbiTaxon": "NCBITaxon_9606",
                "shimada4Class": "unfavorable"
            },
            "bioSamples": {
                "N1": {
                    "pk": 11,
                    "extraInfo": {
                        "inssStage": "2",
                        "lohChr1p": true,
                        "patientStatus": "Relapse_progression",
                        "isTumor": false
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-N1-DNA1-WES1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        }
                    }
                },
                "T1": {
                    "pk": 14,
                    "extraInfo": {
                        "isTumor": true,
                        "inssStage": "2",
                        "lohChr1p": true,
                        "patientStatus": "Relapse_progression"
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 15,
                            "extraInfo": {
                                "extractionType": "DNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 16,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-DNA1-WES1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        },
                        "DNA2": {
                            "pk": 17,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WES2": {
                                    "pk": 18,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T1-RNA1-RNAseq1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": false
                                    }
                                }
                            }
                        }
                    }
                },
                "T2": {
                    "pk": 19,
                    "extraInfo": {
                        "isTumor": true,
                        "inssStage": "3",
                        "lohChr1p": true,
                        "patientStatus": "Relapse_progression"
                    },
                    "testSamples": {
                        "DNA1": {
                            "pk": 20,
                            "extraInfo": {
                                "extractionType": "DNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "WES1": {
                                    "pk": 21,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-DNA1-WES1",
                                        "libraryType": "WES",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        },
                        "RNA1": {
                            "pk": 22,
                            "extraInfo": {
                                "extractionType": "RNA",
                                "testSampleNice": true
                            },
                            "ngsLibraries": {
                                "mRNA_seq1": {
                                    "pk": 23,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "P001-T2-RNA1-mRNAseq1",
                                        "libraryType": "mRNA_seq",
                                        "ngsLibraryNice": true
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()


def test_read_cancer_sheet_custom_fields_header(tsv_sheet_cancer_header):
    sheet = io_tsv.read_cancer_tsv_sheet(tsv_sheet_cancer_header)
    assert EXPECTED_CANCER_SHEET_JSON_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_cancer_sheet_custom_fields_no_header(tsv_sheet_cancer_no_header):
    with pytest.raises(io_tsv.TSVSheetException) as e_info:
        io_tsv.read_cancer_tsv_sheet(tsv_sheet_cancer_no_header)
    EXPECTED = (
        'Unexpected column seen in header row of body: inssStage, lohChr1p, '
        'mycnCn, ncbiTaxon, ngsLibraryNice, patientStatus, shimada4Class, '
        'testSampleNice, timeToDeathDays, timeToEventDays')
    assert EXPECTED == str(e_info.value)


def test_read_cancer_json_custom_fields_header(tsv_sheet_cancer_header):
    sheet_struc = io_tsv.read_cancer_tsv_json_data(tsv_sheet_cancer_header)
    assert EXPECTED_CANCER_SHEET_JSON_HEADER == json.dumps(
        sheet_struc, indent='    ')


def test_read_cancer_json_custom_fields_no_header(tsv_sheet_cancer_no_header):
    with pytest.raises(io_tsv.TSVSheetException) as e_info:
        io_tsv.read_cancer_tsv_json_data(tsv_sheet_cancer_no_header)
    EXPECTED = (
        'Unexpected column seen in header row of body: inssStage, lohChr1p, '
        'mycnCn, ncbiTaxon, ngsLibraryNice, patientStatus, shimada4Class, '
        'testSampleNice, timeToDeathDays, timeToEventDays')
    assert EXPECTED == str(e_info.value)



# -*- coding: utf-8 -*-
"""Tests for loading the compressed germline TSV file
"""

import io
import json
import textwrap

import pytest

from biomedsheets import io_tsv

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def tsv_sheet_germline_header():
    f = io.StringIO(textwrap.dedent("""
    [Metadata]
    schema\tgermline_variants
    schema_version\tv1
    title\tExample germline study
    description\tThe study has two patients, P001 has one tumor sample, P002 has two

    [Data]
    patientName\tfatherName\tmotherName\tsex\tisAffected\tlibraryType\tfolderName\thpoTerms
    12_345\t12_346\t12_347\tM\tY\tWGS\t12_345\tHP:0009946,HP:0009899
    12_348\t12_346\t12_347\tM\tN\tWGS\t12_348\t.
    12_346\t.\t.\tM\tN\t.\t.\t.
    12_347\t.\t.\tF\tN\tWGS\t12_347\t.
    """.lstrip()))
    return f


@pytest.fixture
def tsv_sheet_germline_no_header():
    """Germline TSV sheet without header"""
    f = io.StringIO(textwrap.dedent("""
    patientName\tfatherName\tmotherName\tsex\tisAffected\tlibraryType\tfolderName\thpoTerms
    12_345\t12_346\t12_347\tM\tY\tWGS\t12_345\tHP:0009946,HP:0009899
    12_348\t12_346\t12_347\tM\tN\tWGS\t12_348\t.
    12_346\t.\t.\tM\tN\t.\t.\t.
    12_347\t.\t.\tF\tN\tWGS\t12_347\t.
    """.lstrip()))
    return f


@pytest.fixture
def tsv_sheet_germline_platform_name():
    """Germline TSV sheet with seqPlatform name"""
    f = io.StringIO(textwrap.dedent("""
    patientName\tfatherName\tmotherName\tsex\tisAffected\tlibraryType\tfolderName\thpoTerms\tseqPlatform
    12_347\t.\t.\tF\tN\tWGS\t12_347\t.\tIllumina
    12_347\t.\t.\tF\tN\tWGS\t12_347\t.\tPacBio
    """.lstrip()))
    return f


# Expected value for the germline sheet JSON with header
EXPECTED_GERMLINE_SHEET_JSON_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Example germline study",
    "description": "The study has two patients, P001 has one tumor sample, P002 has two",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            },
            "fatherPk": {
                "docs": "Primary key of mother",
                "key": "fatherPk",
                "type": "string"
            },
            "motherPk": {
                "docs": "Primary key of mother",
                "key": "motherPk",
                "type": "string"
            },
            "fatherName": {
                "docs": "secondary_id of father, used for construction only",
                "key": "fatherName",
                "type": "string"
            },
            "motherName": {
                "key": "motherName",
                "docs": "secondary_id of mother, used for construction only",
                "type": "string"
            },
            "sex": {
                "docs": "Biological sex of individual",
                "key": "sex",
                "type": "enum",
                "choices": [
                    "male",
                    "female",
                    "unknown"
                ]
            },
            "isAffected": {
                "docs": "Flag for marking individiual as (un-)affected",
                "key": "isAffected",
                "type": "enum",
                "choices": [
                    "affected",
                    "unaffected",
                    "unknown"
                ]
            },
            "hpoTerms": {
                "docs": "HPO terms for individual",
                "key": "hpoTerms",
                "type": "array",
                "entry": "string",
                "pattern": "^HPO:[0-9]+$"
            }
        },
        "bioSample": {},
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            }
        }
    },
    "bioEntities": {
        "12_345": {
            "pk": 1,
            "extraInfo": {
                "fatherName": "12_346",
                "motherName": "12_347",
                "sex": "male",
                "isAffected": "affected",
                "hpoTerms": [
                    "HP:0009946",
                    "HP:0009899"
                ],
                "ncbiTaxon": "NCBITaxon_9606",
                "fatherPk": 9,
                "motherPk": 10
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_345",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "12_348": {
            "pk": 5,
            "extraInfo": {
                "fatherName": "12_346",
                "motherName": "12_347",
                "sex": "male",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606",
                "fatherPk": 9,
                "motherPk": 10
            },
            "bioSamples": {
                "N1": {
                    "pk": 6,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 7,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 8,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_348",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "12_346": {
            "pk": 9,
            "extraInfo": {
                "sex": "male",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {}
        },
        "12_347": {
            "pk": 10,
            "extraInfo": {
                "sex": "female",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 11,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_347",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()

# Expected value for the germline sheet JSON without header
EXPECTED_GERMLINE_SHEET_JSON_NO_HEADER = r"""
{
    "identifier": "file://<unknown>",
    "title": "Germline Sample Sheet",
    "description": "Sample Sheet constructed from germline compact TSV file",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            },
            "fatherPk": {
                "docs": "Primary key of mother",
                "key": "fatherPk",
                "type": "string"
            },
            "motherPk": {
                "docs": "Primary key of mother",
                "key": "motherPk",
                "type": "string"
            },
            "fatherName": {
                "docs": "secondary_id of father, used for construction only",
                "key": "fatherName",
                "type": "string"
            },
            "motherName": {
                "key": "motherName",
                "docs": "secondary_id of mother, used for construction only",
                "type": "string"
            },
            "sex": {
                "docs": "Biological sex of individual",
                "key": "sex",
                "type": "enum",
                "choices": [
                    "male",
                    "female",
                    "unknown"
                ]
            },
            "isAffected": {
                "docs": "Flag for marking individiual as (un-)affected",
                "key": "isAffected",
                "type": "enum",
                "choices": [
                    "affected",
                    "unaffected",
                    "unknown"
                ]
            },
            "hpoTerms": {
                "docs": "HPO terms for individual",
                "key": "hpoTerms",
                "type": "array",
                "entry": "string",
                "pattern": "^HPO:[0-9]+$"
            }
        },
        "bioSample": {},
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            }
        }
    },
    "bioEntities": {
        "12_345": {
            "pk": 1,
            "extraInfo": {
                "fatherName": "12_346",
                "motherName": "12_347",
                "sex": "male",
                "isAffected": "affected",
                "hpoTerms": [
                    "HP:0009946",
                    "HP:0009899"
                ],
                "ncbiTaxon": "NCBITaxon_9606",
                "fatherPk": 9,
                "motherPk": 10
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_345",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "12_348": {
            "pk": 5,
            "extraInfo": {
                "fatherName": "12_346",
                "motherName": "12_347",
                "sex": "male",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606",
                "fatherPk": 9,
                "motherPk": 10
            },
            "bioSamples": {
                "N1": {
                    "pk": 6,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 7,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 8,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_348",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "12_346": {
            "pk": 9,
            "extraInfo": {
                "sex": "male",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {}
        },
        "12_347": {
            "pk": 10,
            "extraInfo": {
                "sex": "female",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 11,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 12,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 13,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_347",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()


# Expected value when platform name is given
EXPECTED_GERMLINE_SHEET_JSON_PLATFORM_NAME = r"""
{
    "identifier": "file://<unknown>",
    "title": "Germline Sample Sheet",
    "description": "Sample Sheet constructed from germline compact TSV file",
    "extraInfoDefs": {
        "bioEntity": {
            "ncbiTaxon": {
                "docs": "Reference to NCBI taxonomy",
                "key": "taxon",
                "type": "string",
                "pattern": "^NCBITaxon_[1-9][0-9]*$"
            },
            "fatherPk": {
                "docs": "Primary key of mother",
                "key": "fatherPk",
                "type": "string"
            },
            "motherPk": {
                "docs": "Primary key of mother",
                "key": "motherPk",
                "type": "string"
            },
            "fatherName": {
                "docs": "secondary_id of father, used for construction only",
                "key": "fatherName",
                "type": "string"
            },
            "motherName": {
                "key": "motherName",
                "docs": "secondary_id of mother, used for construction only",
                "type": "string"
            },
            "sex": {
                "docs": "Biological sex of individual",
                "key": "sex",
                "type": "enum",
                "choices": [
                    "male",
                    "female",
                    "unknown"
                ]
            },
            "isAffected": {
                "docs": "Flag for marking individiual as (un-)affected",
                "key": "isAffected",
                "type": "enum",
                "choices": [
                    "affected",
                    "unaffected",
                    "unknown"
                ]
            },
            "hpoTerms": {
                "docs": "HPO terms for individual",
                "key": "hpoTerms",
                "type": "array",
                "entry": "string",
                "pattern": "^HPO:[0-9]+$"
            }
        },
        "bioSample": {},
        "testSample": {
            "extractionType": {
                "docs": "Describes extracted",
                "key": "extractionType",
                "type": "enum",
                "choices": [
                    "DNA",
                    "RNA",
                    "other"
                ]
            }
        },
        "ngsLibrary": {
            "seqPlatform": {
                "docs": "Sequencing platform used",
                "key": "kitName",
                "type": "enum",
                "choices": [
                    "Illumina",
                    "PacBio",
                    "other"
                ]
            },
            "libraryType": {
                "docs": "Rough classificiation of the library type",
                "key": "libraryType",
                "type": "enum",
                "choices": [
                    "Panel-seq",
                    "WES",
                    "WGS",
                    "mRNA-seq",
                    "tRNA-seq",
                    "other"
                ]
            },
            "folderName": {
                "docs": "Name of folder with FASTQ files",
                "key": "folderName",
                "type": "string"
            }
        }
    },
    "bioEntities": {
        "12_347": {
            "pk": 1,
            "extraInfo": {
                "sex": "female",
                "isAffected": "unaffected",
                "ncbiTaxon": "NCBITaxon_9606"
            },
            "bioSamples": {
                "N1": {
                    "pk": 2,
                    "extraInfo": {},
                    "testSamples": {
                        "DNA1": {
                            "pk": 3,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS1": {
                                    "pk": 4,
                                    "extraInfo": {
                                        "seqPlatform": "Illumina",
                                        "folderName": "12_347",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        },
                        "DNA2": {
                            "pk": 5,
                            "extraInfo": {
                                "extractionType": "DNA"
                            },
                            "ngsLibraries": {
                                "WGS2": {
                                    "pk": 6,
                                    "extraInfo": {
                                        "seqPlatform": "PacBio",
                                        "folderName": "12_347",
                                        "libraryType": "WGS"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}""".lstrip()


def test_read_germline_sheet_header(tsv_sheet_germline_header):
    sheet = io_tsv.read_germline_tsv_sheet(tsv_sheet_germline_header)
    assert EXPECTED_GERMLINE_SHEET_JSON_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_germline_sheet_no_header(tsv_sheet_germline_no_header):
    sheet = io_tsv.read_germline_tsv_sheet(tsv_sheet_germline_no_header)
    assert EXPECTED_GERMLINE_SHEET_JSON_NO_HEADER == json.dumps(
        sheet.json_data, indent='    ')


def test_read_germline_sheet_platform_name(tsv_sheet_germline_platform_name):
    sheet = io_tsv.read_germline_tsv_sheet(tsv_sheet_germline_platform_name)
    assert EXPECTED_GERMLINE_SHEET_JSON_PLATFORM_NAME == json.dumps(
        sheet.json_data, indent='    ')


def test_read_tumor_json_header(tsv_sheet_germline_header):
    sheet_struc = io_tsv.read_germline_tsv_json_data(tsv_sheet_germline_header)
    assert EXPECTED_GERMLINE_SHEET_JSON_HEADER == json.dumps(
        sheet_struc, indent='    ')


def test_read_tumor_json_no_header(tsv_sheet_germline_no_header):
    sheet_struc = io_tsv.read_germline_tsv_json_data(tsv_sheet_germline_no_header)
    assert EXPECTED_GERMLINE_SHEET_JSON_NO_HEADER == json.dumps(
        sheet_struc, indent='    ')


def test_read_tumor_json_platform_name(tsv_sheet_germline_platform_name):
    sheet_struc = io_tsv.read_germline_tsv_json_data(tsv_sheet_germline_platform_name)
    assert EXPECTED_GERMLINE_SHEET_JSON_PLATFORM_NAME == json.dumps(
        sheet_struc, indent='    ')



# -*- coding: utf-8 -*-
"""Code for reading and writing the compact germline TSV format
"""

from collections import OrderedDict

from .base import (
    LIBRARY_TYPES, NCBI_TAXON_HUMAN, std_field, TSVSheetException,
    BaseTSVReader)
from ..naming import name_generator_for_scheme, NAMING_DEFAULT

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

#: Default title
GERMLINE_DEFAULT_TITLE = 'Germline Sample Sheet'

#: Default description
GERMLINE_DEFAULT_DESCRIPTION = (
    'Sample Sheet constructed from germline compact TSV file')

#: Germline body TSV header
GERMLINE_BODY_HEADER = ('patientName', 'fatherName', 'motherName', 'sex',
                        'isAffected', 'folderName', 'hpoTerms', 'libraryType')

#: Fixed "extraInfoDefs" field for germline compact TSV
GERMLINE_EXTRA_INFO_DEFS = OrderedDict([
    ('bioEntity', OrderedDict([
        std_field('ncbiTaxon'),
        std_field('fatherPk'),
        std_field('motherPk'),
        ('fatherName', OrderedDict([
            ('docs', 'secondary_id of father, used for construction only'),
            ('key', 'fatherName'),
            ('type', 'string'),
        ])),
        ('motherName', OrderedDict([
            ('key', 'motherName'),
            ('docs', 'secondary_id of mother, used for construction only'),
            ('type', 'string'),
        ])),
        std_field('sex'),
        std_field('isAffected'),
        std_field('hpoTerms'),
    ])),
    ('bioSample', OrderedDict([
    ])),
    ('testSample', OrderedDict([
        std_field('extractionType'),
    ])),
    ('ngsLibrary', OrderedDict([
        std_field('seqPlatform'),
        std_field('libraryType'),
        std_field('folderName'),
    ])),
])

#: Constants for interpreting "sex" field
SEX_VALUES = {
    'M': 'male',
    'm': 'male',
    'f': 'female',
    'F': 'female',
    'U': 'unknown',
    'u': 'unknown',
    '0': 'unknown',
    '1': 'male',
    '2': 'female',
    None: 'unknown,'
}

#: Constants for interpreting "sex" field
AFFECTED_VALUES = {
    'Y': 'affected',
    'y': 'affected',
    'N': 'unaffected',
    'n': 'unaffected',
    'U': 'unknown',
    'u': 'unknown',
    '0': 'unknown',
    '1': 'unaffected',
    '2': 'affected',
    None: 'unknown,'
}


class GermlineTSVSheetException(TSVSheetException):
    """Raised on problems with loading germline TSV sample sheets"""


class GermlineTSVReader(BaseTSVReader):
    """Helper class for reading germline TSV file

    Prefer using ``read_germline_tsv_*()`` for shortcut
    """

    body_header = GERMLINE_BODY_HEADER
    extra_info_defs = GERMLINE_EXTRA_INFO_DEFS
    default_title = GERMLINE_DEFAULT_TITLE
    default_description = GERMLINE_DEFAULT_DESCRIPTION
    bio_entity_name_column = 'patientName'
    bio_sample_name = 'N1'

    def check_tsv_line(self, mapping, lineno):
        """Cancer sample sheet--specific valiation"""
        # Check for hyphen in patient or sample name
        if '-' in mapping['patientName']:
            raise GermlineTSVSheetException(
                'Hyphen not allowed in patientName column')  # pragma: no cover
        if mapping['fatherName'] and '-' in mapping['fatherName']:
            raise GermlineTSVSheetException(
                'Hyphen not allowed in fatherName column')  # pragma: no cover
        if mapping['motherName'] and '-' in mapping['motherName']:
            raise GermlineTSVSheetException(
                'Hyphen not allowed in motherName column')  # pragma: no cover
        # Check "libraryType" field
        if mapping['libraryType'] and (
                mapping['libraryType'] not in LIBRARY_TYPES):
            raise GermlineTSVSheetException(
                'Invalid library type {}, must be in {{{}}}'.format(
                    mapping['libraryType'], ', '.join(LIBRARY_TYPES)))
        # Check "sex" field
        if mapping['sex'] not in SEX_VALUES.keys():
            raise GermlineTSVSheetException(  # pragma: no cover
                ('Invalid "sex" value {} in line {} of data section of '
                 '{}').format(mapping['sex'], lineno + 2, self.fname))
        mapping['sex'] = SEX_VALUES[mapping['sex']]
        # Check "affected"
        if mapping['isAffected'] not in AFFECTED_VALUES.keys():
            raise GermlineTSVSheetException(  # pragma: no cover
                ('Invalid "isAffected" value {} in line {} of data section of '
                 '{}').format(mapping['isAffected'], lineno + 2, self.fname))
        mapping['isAffected'] = AFFECTED_VALUES[mapping['isAffected']]

    def postprocess_json_data(self, json_data):
        """Postprocess JSON data"""
        # Build mapping from bio entity name to pk
        bio_entity_name_to_pk = {
            name: bio_entity['pk']
            for name, bio_entity in json_data['bioEntities'].items()
        }
        # Update bio entities motherPk and fatherPk attributes
        for bio_entity in json_data['bioEntities'].values():
            extra_info = bio_entity['extraInfo']
            if extra_info.get('fatherName', '0') != '0':
                extra_info['fatherPk'] = bio_entity_name_to_pk[
                    extra_info['fatherName']]
            if extra_info.get('motherName', '0') != '0':
                extra_info['motherPk'] = bio_entity_name_to_pk[
                    extra_info['motherName']]
        return json_data

    def construct_bio_entity_dict(self, records, extra_info_defs):
        result = super().construct_bio_entity_dict(records, extra_info_defs)
        result['extraInfo']['ncbiTaxon'] = NCBI_TAXON_HUMAN
        # Check fatherName and motherName entries and assign to result
        self._check_consistency(records, 'fatherName')
        record = records[0]  # representative as it's consistent
        if record.get('fatherName') and record['fatherName'] != '0':
            result['extraInfo']['fatherName'] = record['fatherName']
        if record.get('motherName') and record['motherName'] != '0':
            result['extraInfo']['motherName'] = record['motherName']
        # Check sex and isAffected entries and assign to result
        result['extraInfo']['sex'] = record['sex']
        result['extraInfo']['isAffected'] = record['isAffected']
        # Check hpoTerms entries and assign to result
        if record['hpoTerms']:
            result['extraInfo']['hpoTerms'] = record['hpoTerms'].split(',')
        return result

    @classmethod
    def _check_consistency(cls, records, key):
        values = list(sorted(set(r[key] for r in records)))
        if len(values) > 1:
            raise ValueError(  # pragma: no cover
                'Inconsistent {} entries in records: {}'.format(key, values))


def read_germline_tsv_sheet(f, fname=None, naming_scheme=NAMING_DEFAULT):
    """Read compact germline TSV format from file-like object ``f``

    :return: models.Sheet
    """
    return GermlineTSVReader(f, fname).read_sheet(
        name_generator_for_scheme(naming_scheme))


def read_germline_tsv_json_data(f, fname=None):
    """Read compact germline TSV format from file-like object ``f``

    :return: ``dict``
    """
    return GermlineTSVReader(f, fname).read_json_data()



# -*- coding: utf-8 -*-
"""Tests for the shortcuts module with germline sample sheet"""

import io
import pytest
import textwrap

from biomedsheets import io_tsv, shortcuts


__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def tsv_sheet_germline():
    """Example TSV germline sheet"""
    f = io.StringIO(textwrap.dedent("""
    [Metadata]
    schema\tgermline_variants
    schema_version\tv1
    title\tExample germline study
    description\tSimple study with two trios

    [Data]
    patientName\tfatherName\tmotherName\tsex\tisAffected\tlibraryType\tfolderName\thpoTerms
    index1\tfather1\tmother1\tM\tY\tWES\tindex1\t.
    father1\t0\t0\tM\tN\tWES\tfather1\t.
    mother1\t0\t0\tM\tN\tWES\tmother1\t.
    index2\tfather2\tmother2\tM\tY\tWES\tindex2\t.
    father2\t0\t0\tM\tN\tWES\tfather2\t.
    mother2\t0\t0\tM\tN\tWES\tmother2\t.
    """.lstrip()))
    return f


@pytest.fixture
def sheet_germline(tsv_sheet_germline):
    """Return ``Sheet`` instance for the germline example"""
    return shortcuts.GermlineCaseSheet(io_tsv.read_germline_tsv_sheet(tsv_sheet_germline))


def test_germline_case_sheet(sheet_germline):
    """Tests for the germline case sheet"""
    sheet = sheet_germline
    assert len(sheet.donors) == 6
    assert sheet.cohort
    assert len(sheet.index_ngs_library_to_pedigree) == 2
    assert list(sheet.index_ngs_library_to_pedigree) == [
        'index1-N1-DNA1-WES1-000004', 'index2-N1-DNA1-WES1-000016']
    assert len(sheet.index_ngs_library_to_donor) == 6
    assert list(sheet.library_name_to_library) == [
        'index1-N1-DNA1-WES1-000004',
        'father1-N1-DNA1-WES1-000008',
        'mother1-N1-DNA1-WES1-000012',
        'index2-N1-DNA1-WES1-000016',
        'father2-N1-DNA1-WES1-000020',
        'mother2-N1-DNA1-WES1-000024',
    ]


def test_germline_donor(sheet_germline):
    """Tests for the GermlineDonor objects"""
    index1 = sheet_germline.cohort.indices[0]
    assert index1.father_pk == '5'
    assert index1.father
    assert index1.father.name == 'father1-000005'
    assert index1.mother_pk == '9'
    assert index1.mother
    assert index1.mother.name == 'mother1-000009'


def test_pedigree(sheet_germline):
    """Tests for Pedigree objects"""
    pedigree = sheet_germline.cohort.pedigrees[0]
    assert str(pedigree).startswith('Pedigree(')
    assert [d.name for d in pedigree.donors] == [
        'index1-000001', 'father1-000005', 'mother1-000009']
    assert pedigree.index.name == 'index1-000001'
    assert [d.name for d in pedigree.affecteds] == ['index1-000001']
    assert [d.name for d in pedigree.founders] == ['father1-000005', 'mother1-000009']
    assert list(pedigree.name_to_donor) == ['index1-000001', 'father1-000005', 'mother1-000009']
    assert list(pedigree.pk_to_donor) == ['1', '5', '9']
    assert list(pedigree.secondary_id_to_donor) == ['index1', 'father1', 'mother1']


def test_cohorts(sheet_germline):
    """Tests for Cohort object"""
    cohort = sheet_germline.cohort
    assert len(cohort.pedigrees) == 2
    assert set(d.name for d in cohort.indices) == {'index1-000001', 'index2-000013'}
    assert set(d.name for d in cohort.affecteds) == {'index1-000001', 'index2-000013'}
    assert set(cohort.name_to_pedigree) == {
        'father1-000005',
        'father2-000017',
        'index1-000001',
        'index2-000013',
        'mother1-000009',
        'mother2-000021'}
    assert set(cohort.pk_to_pedigree) == {1, 17, 5, 21, 9, 13}
    assert set(cohort.secondary_id_to_pedigree) == {'index1', 'father1', 'index2', 'mother1', 'father2', 'mother2'}
    assert set(cohort.name_to_donor) == {
        'father1-000005',
        'father2-000017',
        'index1-000001',
        'index2-000013',
        'mother1-000009',
        'mother2-000021'}
    assert set(cohort.secondary_id_to_donor) == {'index1', 'father1', 'mother2', 'index2', 'mother1', 'father2'}
    assert cohort.member_count == 6
    assert cohort.pedigree_count == 2



# -*- coding: utf-8 -*-
"""Base code for reading compact TSV format
"""

from collections import OrderedDict
import re

from .. import ref_resolver
from .. import io
from ..naming import name_generator_for_scheme, NAMING_DEFAULT

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

# TODO: allow setting NCBI taxon through header
# TODO: allow disabling of PK generation through header?

#: Delimiter to use
DELIM = '\t'

#: Known schemas
TSV_SCHEMAS = (
    'cancer_matched', 'germline_variants', 'generic')

#: Known file format versions
FILE_FORMAT_VERSIONS = ('v1',)

#: Keys of TSV [Metadata] section
METADATA_KEYS = ('schema', 'schema_version', 'title', 'description')

#: Valid entities to refer to in custom fields
ENTITY_TYPES = ('bioEntity', 'bioSample', 'testSample', 'ngsLibrary')

#: Valid field types in TSV files
FIELD_TYPES = (
    'string', 'integer', 'boolean', 'number', 'enum', 'pattern')

#: Known library types
LIBRARY_TYPES = (
    'WES', 'WGS', 'Panel_seq', 'mRNA_seq', 'total_RNA_seq')

#: Map library type to extraction type
LIBRARY_TO_EXTRACTION = {
    'WES': 'DNA',
    'WGS': 'DNA',
    'Panel_seq': 'DNA',
    'mRNA_seq': 'RNA',
    'total_RNA_seq': 'RNA',
}

#: Extraction type DNA
EXTRACTION_TYPE_DNA = 'DNA'

#: Extraction type RNA
EXTRACTION_TYPE_RNA = 'RNA'

#: Known extraction types
EXTRACTION_TYPES = (EXTRACTION_TYPE_DNA, EXTRACTION_TYPE_RNA)

#: Key for the TSV header, field "title"
KEY_TITLE = 'title'

#: Key for the TSV header, field "description"
KEY_DESCRIPTION = 'description'

#: Platform "Illumina"
PLATFORM_ILLUMINA = 'Illumina'

#: Platform "PacBio"
PLATFORM_PACBIO = 'PacBio'

#: Platform "other"
PLATFORM_OTHER = 'other'

#: Platform default
PLATFORM_DEFAULT = PLATFORM_ILLUMINA

#: Known platforms
PLATFORM_NAMES = (PLATFORM_ILLUMINA, PLATFORM_PACBIO, PLATFORM_OTHER)

#: Constants for interpreting booleans
BOOL_VALUES = {
    'Y': True,
    'y': True,
    'N': False,
    'n': False,
    'YES': True,
    'yes': True,
    'NO': False,
    'no': False,
    'true': True,
    'false': False,
}

#: NCBI taxon for human
NCBI_TAXON_HUMAN = 'NCBITaxon_9606'


class SheetIOException(Exception):
    """Raised on problems with loading sample sheets"""


class TSVSheetException(SheetIOException):
    """Raised on problems with loading the compact TSV sheets"""


def std_field(name):
    """Return data structure for returning JSON pointer data structure"""
    tpl = ('resource://biomedsheets/data/std_fields.json#'
           '/extraInfoDefs/template/{}')
    return (name, OrderedDict([('$ref', tpl.format(name))]))


class TSVMetadata:
    """Meta data information"""

    def __init__(self, schema=None, schema_version=None, title=None,
                 description=None):
        #: Schema to use for the sample sheet
        self.schema = schema
        #: Version of the sample sheet file format
        self.schema_version = schema_version
        #: Title of the sample sheet
        self.title = title
        #: Description of the sample sheet
        self.description = description


class TSVCustomFieldsInfo:
    """Information about custom fields"""

    def __init__(
            self, key, entity, docs, field_type, minimum, maximum, unit,
            choices, pattern):
        #: Name/key of the field
        self.key = key
        #: The annotated entity
        self.entity = entity
        #: Documentation string, ``None`` if empty
        self.docs = docs
        #: The field type
        self.field_type = field_type
        #: The smallest value
        self.minimum = minimum
        #: The largest value
        self.maximum = maximum
        #: The unit of the value (free text)
        self.unit = unit
        #: Iterable with valid choices
        self.choices = None if choices is None else list(choices)
        #: The regular expression pattern of any
        self.pattern = pattern
        # Ensure all values are valid
        self._validate()

    def _validate(self):
        """Validate the the vaules in this object are valid"""
        if not self.key:
            raise TSVSheetException('Missing field name/key: ' + repr(self.key))
        if self.entity not in ENTITY_TYPES:
            raise TSVSheetException('Invalid entity: ' + repr(self.entity))
        if self.field_type not in FIELD_TYPES:
            raise TSVSheetException(
                'Invalid field type: ' + repr(self.field_type))
        self.minimum = self._convert_minmax(self.minimum, 'minimum')
        self.maximum = self._convert_minmax(self.maximum, 'maximum')
        if self.pattern:
            try:
                re.compile(self.pattern)
            except re.error:
                raise TSVSheetException(
                    'Invalid regular expression ' + repr(self.pattern))

    def _convert_minmax(self, val, which):
        if val is None:
            return None
        else:
            try:
                if self.field_type == 'integer':
                    return int(val)
                else:  # self.field_type == 'number'
                    return float(val)
            except TypeError:
                raise TSVSheetException(
                    'Invalid {} value: {}'.format(which, val))

    def __str__(self):
        return 'TSVCustomFieldsInfo({})'.format(
            ', '.join(map(
                repr, (self.key, self.entity, self.docs, self.field_type,
                       self.minimum, self.maximum, self.unit, self.choices,
                       self.pattern))))

    def __repr__(self):
        return str(self)


class TSVHeader:
    """Header information of succinct TSV files"""

    def __init__(self, metadata=None, custom_field_infos=None):
        #: TSVMetadata with overall information
        self.metadata = metadata
        #: List of TSVCustomFieldsInfo with custom field information
        self.custom_field_infos = OrderedDict([
            (info.key, info) for info in custom_field_infos or []])

    def __str__(self):
        return 'TSVHeader({}, {})'.format(*map(
            repr, (self.metadata, self.custom_field_infos)))

    def __repr__(self):
        return str(self)


#: Header with custom fields
CUSTOM_FIELDS_HEADER = (
    'key\tannotatedEntity\tdocs\ttype\tminimum\tmaximum\tunit\t'
    'choices\tpattern')


class TSVHeaderParser:
    """Parse TSV header from list of header lines"""

    def __init__(self, lines):
        #: List of lines to process
        self.lines = lines

    def run(self):
        """Process header lines and return TSVHeader object"""
        metadata = TSVMetadata()
        custom_fields_header = None
        custom_fields_infos = []
        state = None
        for line in self.lines:
            if line.startswith('#'):
                continue  # comment: ignore
            elif line.startswith('[Metadata]'):
                state = 'metadata'
                continue
            elif line.startswith('[Custom Fields]'):
                state = 'custom_fields'
                continue
            elif line.startswith('[Data]'):
                state = 'data'
                continue
            line = line.strip()
            if not line:
                continue  # skip empty lines
            if state == 'metadata':
                if '\t' not in line:
                    key, value = line, None
                else:
                    key, value = line.split('\t', 1)
                if key not in METADATA_KEYS:
                    raise TSVSheetException('Invalid [Metadata] key: ' + key)
                setattr(metadata, key, value)
            elif state == 'custom_fields':
                if not custom_fields_header:
                    custom_fields_header = line.strip()
                    if custom_fields_header != CUSTOM_FIELDS_HEADER:
                        raise TSVSheetException(
                            ('Invalid custom fields header: {}, '
                             'expected {}').format(
                                 repr(custom_fields_header),
                                 repr(CUSTOM_FIELDS_HEADER)))
                    custom_fields_header = custom_fields_header.split('\t')
                else:
                    custom_fields_infos.append(self._process_custom_fields(
                        line.split('\t')))
        return TSVHeader(metadata, custom_fields_infos)

    @classmethod
    def _process_custom_fields(cls, arr):
        if len(arr) != 9:
            raise TSVSheetException(
                ('Invalid number of entries in [Custom Fields] line: {}, '
                 'should be 9').format(len(arr)))
        arr = [x if x not in ('.', '') else None for x in arr]
        if arr[7]:  # choices
            arr[7] = arr[7].split(',')
        return TSVCustomFieldsInfo(*arr)


class BaseTSVReader:
    """Base class for shortcut TSV readers"""

    #: Iterable with TSV header
    body_header = None
    #: Optional column in body header
    optional_body_header_columns = ('seqPlatform',)
    #: Extra info definitions
    extra_info_defs = None
    #: Default title
    default_title = None
    #: Default description
    default_description = None
    #: Name of bio_entity name column
    bio_entity_name_column = None
    #: Name of the bio_sample name column
    bio_sample_name_column = None
    #: Default sample name if not given in the sample sheet
    bio_sample_name = None
    #: Name of the test_sample name column
    test_sample_name_column = None
    #: Name of the ngs_library name column
    ngs_library_name_column = None
    #: Single extraction type if given
    extraction_type = None

    def __init__(self, f, fname=None):
        self.f = f
        self.fname = fname or '<unknown>'
        self.next_pk = 1

    def read_json_data(self):
        """Read from file-like object ``self.f``, use file name in case of
        problems

        :raises:TSVSheetException in case of problems
        """
        # Read lines from file and check for file not being empty
        lines = [l.strip() for l in self.f]
        if not lines:
            raise TSVSheetException(
                'Problem loading TSV sheet in file {}'.format(
                    self.fname))  # pragma: no cover
        # Decide between the case with or without header
        if lines[0].startswith('['):
            header, body = self._split_lines(lines)
        else:
            header = []
            body = lines
        # Process header and then create a models.Sheet
        tsv_header = TSVHeaderParser(header).run()
        body_header = body[0].split('\t')
        missing_columns = set(self.__class__.body_header) - set(body_header)
        if not body or missing_columns:
            raise TSVSheetException(
                ('Empty or invalid data column names in TSV sheet file {}. '
                 'Must be superset of {{{}}} but is missing {{{}}}').format(
                     self.fname, ', '.join(self.__class__.body_header),
                     ', '.join(missing_columns)))  # pragma: no cover
        # Check for unknown fields
        extra_columns = (
            set(body_header) -
            set(self.__class__.body_header) -
            set(tsv_header.custom_field_infos.keys()) -
            set(self.__class__.optional_body_header_columns))
        if extra_columns:
            msg = 'Unexpected column seen in header row of body: {}'
            raise TSVSheetException(
                msg.format(', '.join(sorted(extra_columns))))
        return self._create_sheet_json(tsv_header, body)

    def read_sheet(self, name_generator=None):
        """Read into JSON and construct ``models.Sheet``"""
        self.name_generator = name_generator or name_generator_for_scheme(
            NAMING_DEFAULT)
        return io.SheetBuilder(self.read_json_data()).run(
            name_generator=name_generator)

    def _split_lines(self, lines):
        """Split string array lines into header and body"""
        header, body = [], []
        in_data = False
        for line in lines:
            if line.startswith('#'):
                continue  # comment, skip
            elif in_data:
                body.append(line)
            else:
                header.append(line)
                if line.startswith('[Data]'):
                    in_data = True
        return header, body

    def _create_sheet_json(self, tsv_header, body):
        """Create models.Sheet object from header dictionary and body lines
        """
        names = body[0].split('\t')  # idx to name
        # Build validated list of records
        records = []
        for lineno, line in enumerate(body[1:]):
            arr = line.split('\t')
            # Replace '.' and '' with ``None``, for empty field
            arr = [(x if x not in ('.', '') else None) for x in arr]
            # Check number of entries in line
            if len(arr) != len(names):
                msg = ('Invalid number of entries in line {} of data '
                       'section of {}: {} vs {}')
                raise TSVSheetException(msg.format(
                    lineno + 2, self.fname, arr, names))  # pragma: no cover
            mapping = dict(zip(names, arr))
            mapping = self.convert_tsv_line(mapping, tsv_header)
            self.check_tsv_line(mapping, lineno)
            records.append(mapping)
        # Create the sheet from records
        return self._create_sheet_json_from_records(tsv_header, records)

    @classmethod
    def convert_tsv_line(cls, mapping, tsv_header):
        """Convert fields in TSV line after conversion to mapping"""
        custom_field_infos = tsv_header.custom_field_infos
        table = {
            'integer': int,
            'number': float,
            'boolean': lambda x: BOOL_VALUES.get(x, False),
        }
        for key, value in mapping.items():
            if value is not None and key in custom_field_infos:
                mapping[key] = table.get(
                    custom_field_infos[key].field_type,
                    lambda x: x)(value)
        return mapping

    def check_tsv_line(self, mapping, lineno):
        """Check TSV line after conversion into mapping

        Override in sub classes
        """
        raise NotImplementedError('Override in sub class')  # pragma: no cover

    def _create_sheet_json_from_records(self, tsv_header, records):
        """Create a new models.Sheet object from TSV records"""
        furl = 'file://{}'.format(self.fname)
        resolver = ref_resolver.RefResolver(dict_class=OrderedDict)
        extra_info_defs = self._augment_extra_info_defs(
            resolver.resolve(furl, self.__class__.extra_info_defs),
            tsv_header)
        json_data = OrderedDict([
            ('identifier', furl),
            ('title', (tsv_header.metadata.title or
                       self.__class__.default_title)),
            ('description', (
                tsv_header.metadata.description or
                self.__class__.default_description)),
            ('extraInfoDefs', extra_info_defs),
            ('bioEntities', OrderedDict()),
        ])
        records_by_bio_entity = OrderedDict()  # records by bio entity
        for record in records:
            records_by_bio_entity.setdefault(
                record[self.__class__.bio_entity_name_column], [])
            records_by_bio_entity[record[
                self.__class__.bio_entity_name_column]].append(record)
        for bio_entity_name, records in records_by_bio_entity.items():
            json_data['bioEntities'][
                bio_entity_name] = self._build_bio_entity_json(
                    records, extra_info_defs)
        return self.postprocess_json_data(json_data)

    @classmethod
    def _augment_extra_info_defs(cls, extra_info_defs, tsv_header):
        """Augment extra definitions with TSV ``[Custom Fields]`` header.
        """
        for info in tsv_header.custom_field_infos.values():
            if info.key not in extra_info_defs[info.entity]:
                extra_def = OrderedDict()
                if info.docs:
                    extra_def['docs'] = info.docs
                extra_def['key'] = info.key
                extra_def['type'] = info.field_type
                if info.pattern:
                    extra_def['pattern'] = info.pattern
                if info.minimum is not None:
                    extra_def['minimum'] = info.minimum
                if info.maximum is not None:
                    extra_def['maximum'] = info.maximum
                if info.choices:
                    extra_def['choices'] = info.choices
                if info.unit:
                    extra_def['unit'] = info.unit
                extra_info_defs[info.entity][info.key] = extra_def
        return extra_info_defs

    def postprocess_json_data(self, json_data):
        """Postprocess JSON data"""
        return json_data

    def _build_bio_entity_json(self, sub_records, extra_info_defs):
        """Build JSON for bio_entities entry"""
        result = self.construct_bio_entity_dict(sub_records, extra_info_defs)
        # If the sub class defines a sample name column then use it to further
        # split up the sample names.  Otherwise, there can only be one sample
        # name and its name must be defined in the class.
        if self.__class__.bio_sample_name_column:
            sample_records = OrderedDict()  # records by bio sample
            for record in sub_records:
                sample_name = record[self.__class__.bio_sample_name_column]
                sample_records.setdefault(sample_name, [])
                sample_records[sample_name].append(record)
            for sample_name, entry in sample_records.items():
                result['bioSamples'][
                    sample_name] = self._build_bio_sample_json(
                        entry, extra_info_defs)
        else:
            assert self.__class__.bio_sample_name, \
                'Must be given if bio_sample_name_column is not'
            # skip if no record with libraryType
            if any(r['libraryType'] for r in sub_records):
                result['bioSamples'][self.__class__.bio_sample_name] = (
                    self._build_bio_sample_json(sub_records, extra_info_defs))
        return result

    def construct_bio_entity_dict(self, records, extra_info_defs):
        """Construct BioEntity dict from bio entity--related record

        Override in sub class to set extraInfo attributes of dict
        """
        self.next_pk += 1
        bio_entity_json = OrderedDict([
            ('pk', self.next_pk - 1),
            ('extraInfo', OrderedDict([
            ])),
            ('bioSamples', OrderedDict()),
        ])
        return self._augment_bio_entity_json(
            bio_entity_json, records[0], extra_info_defs)

    @classmethod
    def _augment_bio_entity_json(
            cls, bio_entity_json, record, extra_info_defs):
        """Augment test sample JSON"""
        for key in extra_info_defs['bioEntity']:
            if key not in bio_entity_json['extraInfo'] and record.get(key):
                bio_entity_json['extraInfo'][key] = record[key]
        return bio_entity_json

    def _build_bio_sample_json(self, records, extra_info_defs):
        """Build JSON for bio_samples entry

        A single test sample entry will be implicitely added.
        """
        self.check_bio_sample_records(records, extra_info_defs)
        result = self.construct_bio_sample_dict(records, extra_info_defs)
        counters_ext = dict((x, 1) for x in EXTRACTION_TYPES)
        counters_lib = dict((x, 1) for x in LIBRARY_TYPES)
        for record in records:
            extraction_type = LIBRARY_TO_EXTRACTION[record['libraryType']]
            if self.__class__.test_sample_name_column:
                test_sample_name = record[
                    self.__class__.test_sample_name_column]
            else:
                test_sample_name = '{}{}'.format(
                    extraction_type, counters_ext[extraction_type])
                counters_ext[extraction_type] += 1
            if self.__class__.ngs_library_name_column:
                lib_name = record[self.__class__.ngs_library_name_column]
            else:
                lib_name = '{}{}'.format(record['libraryType'],
                                         counters_lib[record['libraryType']])
                counters_lib[record['libraryType']] += 1
            pk = self.next_pk
            self.next_pk += 1
            test_sample_json = OrderedDict([
                ('pk', pk),
                ('extraInfo', OrderedDict([
                    ('extractionType', extraction_type),
                ])),
                ('ngsLibraries', OrderedDict([
                    (lib_name, self._build_ngs_library_json(
                        record, extra_info_defs)),
                ]))
            ])
            test_sample_json = self._augment_test_sample_json(
                test_sample_json, record, extra_info_defs)
            result['testSamples'][test_sample_name] = test_sample_json
        return result

    @classmethod
    def _augment_test_sample_json(
            cls, test_sample_json, record, extra_info_defs):
        """Augment test sample JSON"""
        for key in extra_info_defs['testSample']:
            if key not in test_sample_json['extraInfo'] and record.get(key):
                test_sample_json['extraInfo'][key] = record[key]
        return test_sample_json

    def check_bio_sample_records(self, record, extra_info_defs):
        """Override in sub class to check for consistent bio sample records"""

    def construct_bio_sample_dict(self, records, extra_info_defs):
        """Construct BioSample dict from bio sample--related records

        Override in sub class to set extraInfo attributes of dict
        """
        self.next_pk += 1
        bio_sample_json = OrderedDict([
            ('pk', self.next_pk - 1),
            ('extraInfo', OrderedDict()),
            ('testSamples', OrderedDict()),
        ])
        return self._augment_bio_sample_json(
            bio_sample_json, records[0], extra_info_defs)

    @classmethod
    def _augment_bio_sample_json(
            cls, bio_sample_json, record, extra_info_defs):
        """Augment test sample JSON"""
        for key in extra_info_defs['bioSample']:
            if key not in bio_sample_json['extraInfo'] and record.get(key):
                bio_sample_json['extraInfo'][key] = record[key]
        return bio_sample_json

    def _build_ngs_library_json(self, record, extra_info_defs):
        """Build JSON for ngs_libraries entry"""
        result = self.construct_bio_ngs_library_dict(record, extra_info_defs)
        return self._augment_ngs_library_json(result, record, extra_info_defs)

    @classmethod
    def _augment_ngs_library_json(
            cls, ngs_library_json, record, extra_info_defs):
        """Augment NGS library JSON"""
        for key in extra_info_defs['ngsLibrary']:
            if key not in ngs_library_json['extraInfo'] and key in record:
                ngs_library_json['extraInfo'][key] = record[key]
        return ngs_library_json

    def construct_bio_ngs_library_dict(self, record, extra_info_defs):
        """Construct NGSLibrary dict from NGS library--related record

        Override in sub class to set extraInfo attributes of dict
        """
        self.next_pk += 1
        return OrderedDict([
            ('pk', self.next_pk - 1),
            ('extraInfo', OrderedDict([
                ('seqPlatform', record.get('seqPlatform', PLATFORM_DEFAULT)),
                ('folderName', record['folderName']),
                ('libraryType', record['libraryType']),
            ])),
        ])



# -*- coding: utf-8 -*-
"""Tests for the ``__main__`` module

Currently, these are smoke tests only and we should test more stringently
"""

import os

import pytest

from biomedsheets.__main__ import main


@pytest.fixture
def path_json_sheet_cancer():
    """Return path to cancer JSON sheet"""
    return os.path.join(os.path.abspath(
        os.path.dirname(__file__)), 'data', 'example_cancer.json')


@pytest.fixture
def path_tsv_sheet_cancer():
    """Return path to cancer TSV sheet"""
    return os.path.join(os.path.abspath(
        os.path.dirname(__file__)), 'data', 'example_cancer_matched.tsv')


@pytest.fixture
def path_tsv_sheet_germline():
    """Return path to germline TSV sheet"""
    return os.path.join(os.path.abspath(
        os.path.dirname(__file__)), 'data', 'example_germline_variants.tsv')


def test_validate_cancer_json(path_json_sheet_cancer):
    assert not main(['validate', '-i', path_json_sheet_cancer])


def test_expand_cancer_json(path_json_sheet_cancer):
    assert not main(['expand', '-i', path_json_sheet_cancer])


def test_convert_cancer_tsv(path_tsv_sheet_cancer):
    assert not main(
        ['convert', '-t', 'cancer_matched', '-i', path_tsv_sheet_cancer])


def test_convert_germline_tsv(path_tsv_sheet_germline):
    assert not main(
        ['convert', '-t', 'germline_variants', '-i', path_tsv_sheet_germline])



# -*- coding: utf-8 -*-
"""Code for resolving references in JSON code"""

from collections.abc import MutableSequence, MutableMapping
import os
from urllib.parse import urlparse
import sys

import jsonpath_rw
import requests
from requests.exceptions import HTTPError
import requests_file

try:
    import ruamel.yaml as yaml
    YAML_AVAILABLE = True
except ImportError:
    YAML_AVAILABLE = False

from . import requests_resource

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


class RefResolutionException(Exception):
    """Raised on problems with resolving JSON pointers"""


class RefResolver:
    """Helper class for resolving JSON pointers in "$ref" properties

    The resolver will perform on inline updates but rather create new
    dicts and lists with copies.
    """

    def __init__(self, lookup_paths=None, dict_class=dict, verbose=False):
        self.cache = {}
        self.dict_class = dict_class
        self.lookup_paths = list(lookup_paths or [])
        self.session = requests.Session()
        self.session.mount('file://', requests_file.FileAdapter())
        self.session.mount('resource://', requests_resource.ResourceAdapter())
        #: whether or not to resolve relative paths using cwd
        self.rel_cwd_paths = True
        self.verbose = verbose  # TODO: use logging instead

    def resolve(self, doc_uri, obj):
        """Entry point for resolving JSON pointers

        doc_uri is the URI of the JSON document in obj.

        obj can either be a dict or list object (possibly of a sub class)

        raises RefResolutionError on problems with the resolution
        """
        self.cache = {doc_uri: obj}
        return self._resolve(type(obj)(), obj)

    def _resolve(self, base_obj, obj):
        if isinstance(base_obj, (int, bool, float, str)):  # JSON atomic
            return base_obj
        elif isinstance(base_obj, (dict, MutableMapping)):  # JSON object
            return self._resolve_dict_entry(base_obj, obj)
        elif isinstance(base_obj, (list, MutableSequence)):  # JSON list
            return [self._resolve(elem, type(elem)()) for elem in base_obj]
        else:
            raise RefResolutionException(
                'Can only resolve in dict and list container objects and '
                'the atomic types int, bool, float, and str.')

    def _resolve_dict_entry(self, base_obj, obj):
        """Implementation of a dict objects"""
        # Interpret '$ref' key if present in obj, continue to resolve
        # recursively if necessary.
        if base_obj:
            objs = [self._resolve_dict_entry(self.dict_class(), base_obj), obj]
        else:
            objs = [obj]
        imported = set()
        while '$ref' in objs[-1]:
            last = objs[-1]
            if last['$ref'] in imported:
                raise RefResolutionException(
                    'Detected recursion when including {}'.format(
                        last['$ref']))
            imported.add(last['$ref'])
            objs.append(self._load_ref(last['$ref']))
        # Merge objs, values on the left have higher precedence.
        result = self.dict_class()
        for obj in reversed(objs):
            for k, v in obj.items():
                if k != '$ref':
                    if k in result:
                        result[k] = self._resolve(v, result[k])
                    else:
                        result[k] = self._resolve(v, type(v)())
        return result

    def _load_ref(self, ref_uri):
        """Resolve "$ref" URI ``uri``"""
        if self.verbose:
            print('Resolving $ref URI {}'.format(ref_uri), file=sys.stderr)
        parsed_ref_uri = self._parse_ref_uri(ref_uri)
        ref_file = parsed_ref_uri.netloc + parsed_ref_uri.path
        if not ref_file:  # must be relative to current doc
            pass  # is already in cache
        elif ref_file not in self.cache:
            self.cache[ref_uri] = self._load_for_cache(parsed_ref_uri)
        ref_json = self.cache[ref_uri]
        expr = jsonpath_rw.parse(
            '$' + '.'.join(parsed_ref_uri.fragment.split('/')))
        for match in expr.find(ref_json):
            return match.value  # return first match only
        # If we reach here, resolution failed
        raise RefResolutionException(
            'Could not resolve reference URI "{}"'.format(ref_uri))

    def _parse_ref_uri(self, ref_uri):
        parsed_ref_uri = urlparse(ref_uri)
        if (self.rel_cwd_paths and parsed_ref_uri.scheme == 'file' and
                parsed_ref_uri.netloc):
            for path in self.lookup_paths:
                tmp = ref_uri[:(len('file://'))]
                abspath = os.path.join(
                    os.path.abspath(path), parsed_ref_uri.netloc)
                if os.path.exists(abspath):
                    tmp += abspath
                    tmp += ref_uri[len('file://' + parsed_ref_uri.netloc):]
                    return urlparse(tmp)
            else:
                raise RefResolutionException(
                    'Could not find local file {}'.format(
                        parsed_ref_uri.netloc))
        else:
            return parsed_ref_uri

    def _load_for_cache(self, parsed_uri):
        """Load fragment from file URI without cache"""
        remote_uri = '{}://{}/{}'.format(
            parsed_uri.scheme, parsed_uri.netloc, parsed_uri.path)
        if self.verbose:
            print('Loading URI {}'.format(remote_uri), file=sys.stderr)
        response = self.session.get(remote_uri)
        try:
            response.raise_for_status()
        except HTTPError as e:
            raise RefResolutionException(
                'Could not load file {}'.format(parsed_uri.geturl()))
        remote_json = self._load_json(response)
        return remote_json

    def _load_json(self, response):
        if YAML_AVAILABLE:
            return yaml.round_trip_load(response.text)
        else:
            return response.json(object_pairs_hook=self.dict_class)



# -*- coding: utf-8 -*-
"""Tests for the reference resolving code."""

import json
import os

from biomedsheets.ref_resolver import RefResolver


def test_included():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
    resolver = RefResolver([path])
    path_base = path + '/base.json'
    with open(path_base, 'rt') as file_json:
        data_base = json.load(file_json)
    result = resolver.resolve('file://' + path_base, data_base)
    expected = {
        'key1': 'value.base',
        'key2': 'value.included1',
        'key3': 'value.included2',
        'nested': {
            'key1': 1,
            'key2': 2,
            'key3': 3,
        },
        'more': {
            'key': 'value',
        },
    }
    assert expected == result



# -*- coding: utf-8 -*-
"""Code for reading and writing the compact cancer TSV format
"""

from collections import OrderedDict

from .base import (
    LIBRARY_TYPES, BOOL_VALUES, NCBI_TAXON_HUMAN, std_field,
    TSVSheetException, BaseTSVReader)
from ..naming import name_generator_for_scheme, NAMING_DEFAULT

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

#: Default title
CANCER_DEFAULT_TITLE = 'Cancer Sample Sheet'

#: Default description
CANCER_DEFAULT_DESCRIPTION = (
    'Sample Sheet constructed from cancer matched samples '
    'compact TSV file')

#: Cancer body TSV header
CANCER_BODY_HEADER = (
    'patientName', 'sampleName', 'isTumor', 'libraryType', 'folderName')

#: Fixed "extraInfoDefs" field for cancer compact TSV
CANCER_EXTRA_INFO_DEFS = OrderedDict([
    ('bioEntity', OrderedDict([
        std_field('ncbiTaxon'),
    ])),
    ('bioSample', OrderedDict([
        std_field('isTumor'),
    ])),
    ('testSample', OrderedDict([
        std_field('extractionType'),
    ])),
    ('ngsLibrary', OrderedDict([
        std_field('seqPlatform'),
        std_field('libraryType'),
        std_field('folderName'),
    ])),
])


class CancerTSVSheetException(TSVSheetException):
    """Raised on problems with loading cancer TSV sample sheets"""


class CancerTSVReader(BaseTSVReader):
    """Helper class for reading cancer TSV file

    Prefer using ``read_cancer_tsv_*()`` for shortcut
    """

    body_header = CANCER_BODY_HEADER
    extra_info_defs = CANCER_EXTRA_INFO_DEFS
    default_title = CANCER_DEFAULT_TITLE
    default_description = CANCER_DEFAULT_DESCRIPTION
    bio_entity_name_column = 'patientName'
    bio_sample_name_column = 'sampleName'

    def check_tsv_line(self, mapping, lineno):
        """Cancer sample sheet--specific valiation"""
        # Check for hyphen in patient or sample name
        if '-' in mapping['patientName']:
            raise CancerTSVSheetException(
                'Hyphen not allowed in patientName column')  # pragma: no cover
        if '-' in mapping['sampleName']:
            raise CancerTSVSheetException(
                'Hyphen not allowed in sampleName column')  # pragma: no cover
        # Check "isTumor" field, convert to bool
        if mapping['isTumor'] not in BOOL_VALUES.keys():
            raise CancerTSVSheetException(  # pragma: no cover
                ('Invalid boolean value {} in line {} '
                 'of data section of {}').format(
                     mapping['isTumor'], lineno + 2, self.fname))
        mapping['isTumor'] = BOOL_VALUES[mapping['isTumor']]
        # Check "libraryType" field
        if mapping['libraryType'] not in LIBRARY_TYPES:
            raise CancerTSVSheetException(  # pragma: no cover
                'Invalid library type {}, must be in {{{}}}'.format(
                    mapping['libraryType'], ', '.join(LIBRARY_TYPES)))
        # Check other fields for being non-empty
        for key in self.__class__.body_header:
            if mapping[key] is None:
                raise CancerTSVSheetException(
                    'Field {} empty in line {} of {}'.format(
                        key, lineno + 2, self.fname))  # pragma: no cover
        # TODO: we should perform more validation here in the future

    def construct_bio_entity_dict(self, records, extra_info_defs):
        result = super().construct_bio_entity_dict(records, extra_info_defs)
        result['extraInfo']['ncbiTaxon'] = NCBI_TAXON_HUMAN
        return result

    def check_bio_sample_records(self, records, extra_info_defs):
        if len(set(r['isTumor'] for r in records)) != 1:
            raise TSVSheetException(
                'Inconsistent "isTumor" flag for records')  # pragma: no cover

    def construct_bio_sample_dict(self, records, extra_info_defs):
        result = super().construct_bio_sample_dict(records, extra_info_defs)
        result['extraInfo']['isTumor'] = records[0]['isTumor']
        return result


def read_cancer_tsv_sheet(f, fname=None, naming_scheme=NAMING_DEFAULT):
    """Read compact cancer TSV format from file-like object ``f``

    :return: models.Sheet
    """
    return CancerTSVReader(f, fname).read_sheet(
        name_generator_for_scheme(naming_scheme))


def read_cancer_tsv_json_data(f, fname=None):
    """Read compact cancer TSV format from file-like object ``f``

    :return: ``dict``
    """
    return CancerTSVReader(f, fname).read_json_data()



# -*- coding: utf-8 -*-
"""Extensions for requests module"""

import io
import locale
import pkg_resources

from requests.adapters import BaseAdapter
from requests.compat import urlparse
from requests import Response, codes


class ResourceAdapter(BaseAdapter):
    """Adapter to pkg_resources.resource_stream()"""

    def send(self, request, **kwargs):
        """Wraps a pkg_resource.resource_stream

        The host name is interpreted as the package name
        """
        # Check that the method makes sense. Only support GET
        if request.method not in ('GET', 'HEAD'):
            raise ValueError('Invalid request method {}'.format(
                request.method))
        # Parse the URL
        url_parts = urlparse(request.url)

        # Interpret host name as package name
        if not url_parts.netloc:
            raise ValueError(
                'pkg_resource: hostname interpreted as package name')
        pkg_name = url_parts.netloc

        resp = Response()

        # Open the resource stream, translate certain errors into HTTP
        # responses. Use urllib's unquote to translate percent escapes into
        # whatever they actually need to be
        try:
            resp.raw = pkg_resources.resource_stream(
                pkg_name, url_parts.path)
            resp.raw.release_conn = resp.raw.close
        except FileNotFoundError as e:
            resp.status_code = codes.not_found

            # Wrap the error message in a file-like object
            # The error message will be localized, try to convert the string
            # representation of the exception into a byte stream
            resp_str = str(e).encode(locale.getpreferredencoding(False))
            resp.raw = io.BytesIO(resp_str)
            resp.headers['Content-Length'] = len(resp_str)

            # Add release_conn to the BytesIO object
            resp.raw.release_conn = resp.raw.close
        else:
            resp.status_code = codes.ok

        return resp

    def close(self):
        pass



# -*- coding: utf-8 -*
"""Shortcuts for rare germline sample sheets
"""

from collections import OrderedDict
from warnings import warn

from .base import (
    EXTRACTION_TYPE_DNA, EXTRACTION_TYPE_RNA, MissingDataEntity,
    ShortcutSampleSheet, BioSampleShortcut, TestSampleShortcut,
    NGSLibraryShortcut, MissingDataWarning)
from .generic import GenericBioEntity
from ..union_find import UnionFind

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'

# TODO: need selector or WGS/WES data somehow in case both are present.

#: Key value for "extraction type" value
KEY_EXTRACTION_TYPE = 'extractionType'

#: Key value for the "is affected" flag
KEY_IS_AFFECTED = 'isAffected'

#: Key value for the "father PK" value
KEY_FATHER_PK = 'fatherPk'

#: Key value for the "mother PK" value
KEY_MOTHER_PK = 'motherPk'

#: Key value for "sex".
KEY_SEX = 'sex'


class Pedigree:
    """Class for accessing information in a pedigree

    The individuals are represented by :py:class:`GermlineDonor` objects.

    Note that the shortcut members are set upon creation.  When members are
    modified after the construction, the ``update_shortcuts()`` method must
    be called.
    """

    def __init__(self, donors=None, index=None):
        #: Members of the pedigree
        self.donors = list(donors or [])
        #: Index patient in the pedigree, there can only be one, even if
        #: there are multiple affected individuals.  Usually, the first
        #: affected patient in a study is used
        self.index = index
        #: All affected individuals
        self.affecteds = []
        #: Founders in the pedigree, assumed to be unrelated
        self.founders = []
        #: Mapping from individual name to donor individual
        self.name_to_donor = {}
        #: Mapping from individual pk to donor individual
        self.pk_to_donor = {}
        #: Mapping from individual secondary_id to donor individual
        self.secondary_id_to_donor = {}
        # Initialize the shortcuts
        self.update_shortcuts()

    @property
    def member_count(self):
        """Return number of members in the pedigree"""
        return len(self.donors)

    def update_shortcuts(self):
        """Update the shortcut members"""
        if len(self.donors) == 1:
            # For singletons, use the single individual as index regardless
            # of affection state.  This allows the usage of cancer sample
            # sheets in variant_calling.
            self.affecteds = []
            self.index = self.donors[0]
        else:
            self.affecteds = [d for d in self.donors if d.is_affected]
            if self.index is None and self.affecteds:
                self.index = self.affecteds[0]
            elif self.index is None:  # fallback to first in list
                self.index = self.donors[0]
        self.founders = [d for d in self.donors if d.is_founder]
        self.name_to_donor = OrderedDict([(d.name, d) for d in self.donors])
        self.pk_to_donor = OrderedDict([(str(d.pk), d) for d in self.donors])
        self.secondary_id_to_donor = OrderedDict([
            (d.secondary_id, d) for d in self.donors])

    def __repr__(self):
        return 'Pedigree({})'.format(', '.join(map(str, [
            self.donors, self.index])))

    def __str__(self):
        return repr(self)


def _append_pedigree_to_ped(pedigree, f):
    family = 'FAM_' + pedigree.index.name
    for donor in pedigree.donors:
        affected = {
            'affected': '2', 'unaffected': '1', 'unknown': '0'
        }[donor.extra_infos.get(KEY_IS_AFFECTED, 'unknown')]
        sex = {'male': '1', 'female': '2', 'unknown': '0'}[
            donor.extra_infos.get(KEY_SEX, 'unknown')]
        father = '0'
        if donor.father_pk:
            if hasattr(pedigree.pk_to_donor[donor.father_pk],
                       'dna_ngs_library'):
                donor_father = pedigree.pk_to_donor[donor.father_pk]
                if donor_father.dna_ngs_library is None:
                    father = donor_father.name
                else:
                    father = donor_father.dna_ngs_library.name
        mother = '0'
        if donor.mother_pk:
            if hasattr(pedigree.pk_to_donor[donor.mother_pk],
                       'dna_ngs_library'):
                donor_mother = pedigree.pk_to_donor[donor.mother_pk]
                if donor_mother.dna_ngs_library is None:
                    mother = donor_mother.name
                else:
                    mother = donor_mother.dna_ngs_library.name
        if hasattr(donor, 'dna_ngs_library'):
            if donor.dna_ngs_library is None:
                name = donor.name
            else:
                name = donor.dna_ngs_library.name
            print('\t'.join(
                (family, name, father, mother, sex, affected)), file=f)


def write_pedigree_to_ped(pedigree, path):
    with open(path, 'wt') as f:
        _append_pedigree_to_ped(pedigree, f)


def write_pedigrees_to_ped(pedigrees, path):
    with open(path, 'wt') as f:
        for pedigree in pedigrees:
            _append_pedigree_to_ped(pedigree, f)


class Cohort:
    """Class for accessing information about a set of :py:class:`Pedigree`:
    objects.

    Pedigrees are assumed to not overlap.

    Note that the shortcut members are set upon creation.  When pedigrees are
    modified after the construction, the ``update_shortcuts()`` method must
    be called.
    """

    def __init__(self, pedigrees=None):
        #: The pedigrees in the cohort
        self.pedigrees = list(pedigrees or [])
        #: List of all index individuals of all pedigrees
        self.indices = []
        #: List of all affected individuals of all pedigrees
        self.affecteds = []
        #: Mapping from individual name to pedigree
        self.name_to_pedigree = {}
        #: Mapping from individual pk to pedigree
        self.pk_to_pedigree = {}
        #: Mapping from individual secondary_id to pedigree
        self.secondary_id_to_pedigree = {}
        #: Mapping from individual name to donor individual
        self.name_to_donor = {}
        #: Mapping from individual pk to donor individual
        self.pk_to_donor = {}
        #: Mapping from individual secondary_id to donor individual
        self.secondary_id_to_donor = {}
        # Initialize the shortcuts
        self.update_shortcuts()

    @property
    def member_count(self):
        """Return number of members in the cohort"""
        return sum(p.member_count for p in self.pedigrees)

    @property
    def pedigree_count(self):
        """Return number of pedigrees in the cohort"""
        return len(self.pedigrees)

    def update_shortcuts(self):
        """Update the shortcut members"""
        # Re-build lists of index and affected individuals
        self.indices = [p.index for p in self.pedigrees]
        self.affecteds = sum((p.affecteds for p in self.pedigrees), [])
        # Re-build mappings
        self.name_to_pedigree = {}
        self.pk_to_pedigree = {}
        self.secondary_id_to_pedigree = {}
        self.name_to_donor = {}
        self.pk_to_donor = {}
        self.secondary_id_to_donor = {}
        for pedigree in self.pedigrees:
            # Update {name,pk,secondary_id} to pedigree mapping
            self._checked_update(
                self.name_to_pedigree,
                {d.name: pedigree for d in pedigree.donors}, 'name')
            self._checked_update(
                self.pk_to_pedigree,
                {d.pk: pedigree for d in pedigree.donors}, 'pk')
            self._checked_update(
                self.secondary_id_to_pedigree,
                {d.secondary_id: pedigree for d in pedigree.donors},
                'secondary id')
            # Update {name,pk,secondary_id} to donor mapping
            self._checked_update(
                self.name_to_donor,
                {d.name: d for d in pedigree.donors}, 'name')
            self._checked_update(
                self.pk_to_donor, {d.pk: d for d in pedigree.donors}, 'pk')
            self._checked_update(
                self.secondary_id_to_donor, {
                    d.secondary_id: d for d in pedigree.donors},
                'secondary id')

    def _checked_update(self, dest, other, msg_token):
        """Check overlap of keys between ``dest`` and ``other``, update and
        return dest

        In case of error, use ``msg_token`` for exception message.
        """
        overlap = set(dest.keys()) & set(other.keys())
        if overlap:
            tpl = ('Duplicate {}s when building '
                   'cohort shortcuts: {}')  # pramga: no cover
            raise ValueError(  # pramga: no cover
                tpl.format(msg_token, list(sorted(overlap))))
        dest.update(other)
        return dest


class CohortBuilder:
    """Helper class for building a :py:class:`Cohort` object from an iterable
    of :py:class:`GermlineDonor` objects

    Also initialize the internal father and mother attributes of
    :py:class:`GermlineDonor`
    """

    def __init__(self, donors):
        #: Iterable of :py:class:`GermlineDonor` objects
        self.donors = list(donors)

    def run(self):
        """Return :py:class:`Cohort` object with :py:class:`Pedigree` sub
        structure
        """
        cohort = Cohort(self._yield_pedigrees())
        for pedigree in cohort.pedigrees:
            for donor in pedigree.donors:
                if donor.father_pk:
                    donor._father = cohort.pk_to_donor[int(donor.father_pk)]
                if donor.mother_pk:
                    donor._mother = cohort.pk_to_donor[int(donor.mother_pk)]
        return cohort

    def _yield_pedigrees(self):
        """Yield Pedigree objects built from self.donors"""
        # Use Union-Find data structure for gathering pedigree donors
        union_find = UnionFind()
        for donor in self.donors:
            for parent_pk in (
                    pk for pk in (donor.father_pk, donor.mother_pk) if pk):
                # String conversion is necessary because "fatherPk" and
                # "motherPk" are given with type "str" in std_fields.json
                union_find.union(str(donor.pk), parent_pk)
        # Partition the donors
        partition = OrderedDict()
        for donor in self.donors:
            partition.setdefault(union_find[str(donor.pk)], []).append(donor)
        # Construct the pedigrees
        for ped_donors in partition.values():
            affecteds = [d for d in ped_donors if d.is_affected]
            yield Pedigree(ped_donors)


class GermlineDonor(GenericBioEntity):
    """Represent a donor in a germline study"""

    def __init__(self, shortcut_sheet, bio_entity):
        super().__init__(shortcut_sheet, bio_entity)
        # ``GermlineDonor`` object for father, access via property, set in
        # ``CohortBuilder``
        self._father = None
        # ``GermlineDonor`` object for mother, access via property, set in
        # ``CohortBuilder``
        self._mother = None
        #: The primary bio sample with DNA
        self.dna_bio_sample = self._get_primary_dna_bio_sample()
        #: The primary bio sample with RNA, if any
        self.rna_bio_sample = self._get_primary_rna_bio_sample()
        #: The primary DNA test sample
        self.dna_test_sample = self._get_primary_dna_test_sample()
        #: The primary RNA test sample, if any
        self.rna_test_sample = self._get_primary_rna_test_sample()
        #: The primary DNA NGS library for this sample
        self.dna_ngs_library = self._get_primary_dna_ngs_library()
        #: The primary RNA NGS library for this sample, if any
        self.rna_ngs_library = self._get_primary_rna_ngs_library()

    @property
    def is_affected(self):
        """Return whether or not the donor is affected"""
        return self.extra_infos.get(
            KEY_IS_AFFECTED, 'unaffected') == 'affected'

    @property
    def father_pk(self):
        """Return PK of father or ``None``"""
        return self.extra_infos.get(KEY_FATHER_PK, None)

    @property
    def mother_pk(self):
        """Return PK of mother or ``None``"""
        return self.extra_infos.get(KEY_MOTHER_PK, None)

    @property
    def father(self):
        """Return mother ``GermlineDonor`` object or ``None``"""
        if not self._father and self.father_pk:
            raise AttributeError(
                'Father object not yet set, although PK available.  '
                'Not processed through CohortBuilder?')
        return self._father

    @property
    def mother(self):
        """Return mother ``GermlineDonor`` object or ``None``"""
        if not self._mother and self.mother_pk:
            raise AttributeError(
                'Mother object not yet set, although PK available.  '
                'Not processed through CohortBuilder?')
        return self._mother

    @property
    def is_founder(self):
        """Return whether is founder, i.e., has neither mother nor father"""
        return (not self.father_pk) and (not self.mother_pk)

    def _get_primary_dna_bio_sample(self):
        """Spider through ``self.bio_entity`` and return primary bio sample
        """
        for sample in self._iter_all_bio_samples(EXTRACTION_TYPE_DNA, True):
            if not sample.extra_infos.get('isTumor', False):
                return BioSampleShortcut(self, sample, 'ngs_library')
        return None

    def _get_primary_rna_bio_sample(self):
        """Spider through ``self.bio_entity`` and return primary bio sample
        with RNA data, if any; ``None`` otherwise
        """
        for sample in self._iter_all_bio_samples(EXTRACTION_TYPE_RNA, True):
            if not sample.extra_infos.get('isTumor', False):
                return BioSampleShortcut(self, sample, 'ngs_library')
        return None

    def _get_primary_dna_test_sample(self):
        """Spider through ``self.bio_entity`` and return primary DNA test sample
        """
        if (self.dna_bio_sample and
                self.dna_bio_sample.bio_sample.test_samples):
            return TestSampleShortcut(self.dna_bio_sample, next(iter(
                self.dna_bio_sample.bio_sample.test_samples.values())),
                'ngs_library')
        else:
            return None

    def _get_primary_rna_test_sample(self):
        """Spider through ``self.bio_entity`` and return primary RNA
        testsample, if any; ``None`` otherwise
        """
        if (self.rna_bio_sample and
                self.rna_bio_sample.bio_sample.test_samples):
            return TestSampleShortcut(self.rna_bio_sample, next(iter(
                self.rna_bio_sample.bio_sample.test_samples.values())),
                'ngs_library')
        else:
            return None

    def _get_primary_dna_ngs_library(self):
        """Get primary DNA NGS library from self.dna_test_sample
        """
        if (self.dna_test_sample and
                self.dna_test_sample.test_sample.ngs_libraries):
            return NGSLibraryShortcut(self.dna_test_sample, next(iter(
                self.dna_test_sample.test_sample.ngs_libraries.values())))
        else:
            return None

    def _get_primary_rna_ngs_library(self):
        """Get primary RNA NGS library from self.rna_test_sample, if any
        """
        if (self.rna_test_sample and
                self.rna_test_sample.test_sample.ngs_libraries):
            return NGSLibraryShortcut(self.rna_test_sample, next(iter(
                self.rna_test_sample.test_sample.ngs_libraries.values())))
        else:
            return None

    def _iter_all_bio_samples(self, ext_type, allow_none):
        """Yield all bio samples with a test sample of the given extraction
        type

        Require yielding of at least one unless ``allow_none``
        """
        yielded_any = False
        for bio_sample in self.bio_entity.bio_samples.values():
            for test_sample in bio_sample.test_samples.values():
                if KEY_EXTRACTION_TYPE not in test_sample.extra_infos:
                    raise MissingDataEntity(
                        'Could not find "{}" flag in TestSample {}'.format(
                            KEY_EXTRACTION_TYPE, test_sample))
                elif test_sample.extra_infos[KEY_EXTRACTION_TYPE] == ext_type:
                    yielded_any = True
                    yield bio_sample
                    break  # each bio_sample only once
        if not yielded_any and not allow_none:
            msg = 'Could not find a TestSample with {} == {} for BioEntity {}'
            raise MissingDataEntity(msg.format(
                KEY_EXTRACTION_TYPE, ext_type, self.bio_entity))

    def __repr__(self):
        return 'GermlineDonor({})'.format(', '.join(map(
            str, [self.sheet, self.bio_entity])))

    def __str__(self):
        return repr(self)


class GermlineCaseSheet(ShortcutSampleSheet):
    """Shortcut for "germline" view on bio-medical sample sheets"""

    bio_entity_class = GermlineDonor

    def __init__(self, sheet):
        super().__init__(sheet)
        #: List of donors in the sample sheet
        self.donors = list(self._iter_donors())
        #: :py:class:`Cohort` object with the pedigrees and donors built from
        #: the sample sheet
        self.cohort = CohortBuilder(self.donors).run()
        #: Mapping from index DNA NGS library name to pedigree
        self.index_ngs_library_to_pedigree = OrderedDict(
            self._index_ngs_library_to_pedigree())
        #: Mapping from any DNA NGS library name in pedigree to pedigree
        self.donor_ngs_library_to_pedigree = OrderedDict([
            (donor.dna_ngs_library.name, pedigree)
            for pedigree in self.cohort.pedigrees
            for donor in pedigree.donors if donor.dna_ngs_library])
        #: Mapping from DNA NGS library name to donor
        self.index_ngs_library_to_donor = OrderedDict([
            (donor.dna_ngs_library.name, donor)
            for donor in self.donors if donor.dna_ngs_library])
        #: Mapping from library name to object
        self.library_name_to_library = OrderedDict(
            self._library_name_to_library())

    def _iter_donors(self):
        """Return iterator over the donors in the study"""
        for bio_entity in self.sheet.bio_entities.values():
            yield GermlineDonor(self, bio_entity)

    def _index_ngs_library_to_pedigree(self):
        """Build mapping from NGS library name to pedigree"""
        for pedigree in self.cohort.pedigrees:
            if not pedigree.index:
                raise ValueError(  # pragma: no cover
                    'Found pedigree without index! {}'.format(pedigree))
            if not pedigree.index.dna_ngs_library:
                # Warn if pedigree does not have a NGS library
                tpl = 'Pedigree index has no DNA library! {}/{}'
                msg = tpl.format(pedigree.index, pedigree)
                warn(msg, MissingDataWarning)
                continue
            yield pedigree.index.dna_ngs_library.name, pedigree

    def _library_name_to_library(self):
        """Yield mapping from library name to library"""
        for donor in self.donors:
            for bio_sample in donor.bio_samples.values():
                for test_sample in bio_sample.test_samples.values():
                    for ngs_library in test_sample.ngs_libraries.values():
                        yield ngs_library.name, ngs_library



# -*- coding: utf-8 -*-
"""Shortcuts for cancer sample sheets
"""

from collections import OrderedDict
from warnings import warn

from .base import (
    EXTRACTION_TYPE_DNA, EXTRACTION_TYPE_RNA,
    MissingDataEntity, MissingDataWarning, ShortcutSampleSheet,
    TestSampleShortcut, NGSLibraryShortcut)
from .generic import (GenericBioEntity, GenericBioSample)

__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


#: Key value for "is tumor" flag
KEY_IS_TUMOR = 'isTumor'

#: Key value for "extraction type" value
KEY_EXTRACTION_TYPE = 'extractionType'


class CancerCaseSheetOptions:
    """Options for parsing cancer case sheets"""

    @classmethod
    def create_defaults(klass):
        return CancerCaseSheetOptions(
            allow_missing_normal=False,
            allow_missing_tumor=False)

    def __init__(self, *, allow_missing_normal, allow_missing_tumor):
        #: Allow missing normal sample, leads to empty pair list
        self.allow_missing_normal = allow_missing_normal
        #: Allow missing tumor sample, leads to empty pair list
        self.allow_missing_tumor = allow_missing_tumor

    def __str__(self):
        args = (self.allow_missing_normal, self.allow_missing_tumor)
        return 'CancerCaseSheetOptions({})'.format(', '.join(map(str, args)))

    def __repr__(self):
        return str(self)


class CancerCaseSheet(ShortcutSampleSheet):
    """Shortcut for "matched tumor/normal" view on bio-medical sample sheets

    Note that by default, at least one normal and one tumor sample must be
    given.  This can be changed by setting the ``options`` on creation.
    """

    def __init__(self, sheet, options=None):
        super().__init__(sheet)
        #: Configuration
        self.options = options or CancerCaseSheetOptions.create_defaults()
        #: List of donors in the sample sheet
        self.donors = list(self._iter_donors())
        #: List of primary matched tumor/normal sample pairs in the sheet
        self.primary_sample_pairs = list(self._iter_sample_pairs(True))
        #: List of all matched tumor/normal sample pairs in the sample sheet
        self.all_sample_pairs = list(self._iter_sample_pairs(False))
        #: Mapping of all sample pairs by name of primary DNA test sample
        self.all_sample_pairs_by_tumor_dna_test_sample = OrderedDict(
            (pair.tumor_sample.dna_test_sample.name, pair)
            for pair in self.all_sample_pairs
            if pair.tumor_sample and pair.tumor_sample.dna_test_sample)
        #: Mapping of all sample pairs by name of primary DNA library name
        self.all_sample_pairs_by_tumor_dna_ngs_library = OrderedDict(
            (pair.tumor_sample.dna_ngs_library.name, pair)
            for pair in self.all_sample_pairs
            if pair.tumor_sample.dna_ngs_library)
        #: Mapping of all sample pairs by name of primary RNA library name
        self.all_sample_pairs_by_tumor_rna_ngs_library = OrderedDict(
            (pair.tumor_sample.rna_ngs_library.name, pair)
            for pair in self.all_sample_pairs
            if pair.tumor_sample.rna_ngs_library)

    def _iter_donors(self):
        """Return iterator over the donors in the study"""
        for bio_entity in self.sheet.bio_entities.values():
            yield CancerDonor(self, bio_entity)

    def _iter_sample_pairs(self, only_primary_sample_pairs):
        """Return iterator over the matched tumor/normal pairs

        If ``only_primary_sample`` is ``True`` then only one pair per donor
        will be returned with the cancer sample marked as primary.  If this
        flag is ``False`` then one pair for each ``cancer`` sample will be
        returned.
        """
        for donor in self.donors:
            if only_primary_sample_pairs:
                yield donor.primary_pair
            else:
                yield from donor.all_pairs


class CancerMatchedSamplePair:
    """Represents a matched tumor/normal sample pair"""

    def __init__(self, donor, tumor_sample, normal_sample):
        #: The ``BioEntity`` from the sample sheet
        self.donor = donor
        #: Alias for ``self.donor``
        self.bio_entity = self.donor
        #: The ``CancerBioSample`` from the sample sheet
        self.tumor_sample = tumor_sample
        #: The ``CancerBioSample`` from the sample sheet
        self.normal_sample = normal_sample

    def __repr__(self):
        return 'CancerMatchedSamplePair({})'.format(', '.join(
            map(str, [self.donor, self.tumor_sample, self.normal_sample])))

    def __str__(self):
        return repr(self)


class CancerBioSample(GenericBioSample):
    """Represents one sample in a tumor/normal sample pair in the context
    of a matched Cancer tumor/normal study

    Currently, only NGS samples are supported and at least one DNA library is
    required.  This will change in the future
    """

    def __init__(self, shortcut_bio_entity, bio_sample):
        super().__init__(shortcut_bio_entity, bio_sample)
        #: The ``CancerDonor`` from the sample sheet
        self.donor = shortcut_bio_entity
        #: The primary DNA test sample
        self.dna_test_sample = self._get_primary_dna_test_sample()
        #: The primary RNA test sample, if any
        self.rna_test_sample = self._get_primary_rna_test_sample()
        #: The primary DNA NGS library for this sample
        self.dna_ngs_library = self._get_primary_dna_ngs_library()
        #: The primary RNA NGS library for this sample, if any
        self.rna_ngs_library = self._get_primary_rna_ngs_library()

    @property
    def is_tumor(self):
        """Whether or not the bio sample is cancerous"""
        return self.extra_infos[KEY_IS_TUMOR]

    def _get_primary_dna_test_sample(self):
        """Spider through ``self.bio_sample`` and return primary DNA test
        sample
        """
        sample = next(self._iter_all_test_samples(EXTRACTION_TYPE_DNA, True),
                      None)
        if sample:
            return TestSampleShortcut(self, sample, 'ngs_library')
        else:
            return None

    def _get_primary_rna_test_sample(self):
        """Spider through ``self.bio_sample`` and return primary RNA test
        sample, if any; ``None`` otherwise
        """
        sample = next(self._iter_all_test_samples(EXTRACTION_TYPE_RNA, True),
                      None)
        if sample:
            return TestSampleShortcut(self, sample, 'ngs_library')
        else:
            return None

    def _get_primary_dna_ngs_library(self):
        """Get primary DNA NGS library from self.dna_test_sample
        """
        if (self.dna_test_sample and
                self.dna_test_sample.test_sample.ngs_libraries):
            return NGSLibraryShortcut(self.dna_test_sample, next(iter(
                self.dna_test_sample.test_sample.ngs_libraries.values())))
        else:
            return None

    def _get_primary_rna_ngs_library(self):
        """Get primary RNA NGS library from self.rna_test_sample, if any
        """
        if (self.rna_test_sample and
                self.rna_test_sample.test_sample.ngs_libraries):
            return NGSLibraryShortcut(self.rna_test_sample, next(iter(
                self.rna_test_sample.test_sample.ngs_libraries.values())))
        else:
            return None

    def _iter_all_test_samples(self, ext_type, allow_none):
        """Yield all test samples with the given extraction type

        Require yielding of at least one unless ``allow_none``
        """
        yielded_any = False
        for test_sample in self.bio_sample.test_samples.values():
            if KEY_EXTRACTION_TYPE not in test_sample.extra_infos:
                raise MissingDataEntity(  # pragma: no cover
                    'Could not find "{}" flag in TestSample {}'.format(
                        KEY_EXTRACTION_TYPE, test_sample))
            elif test_sample.extra_infos[KEY_EXTRACTION_TYPE] == ext_type:
                yielded_any = True
                yield test_sample
        if not yielded_any and not allow_none:
            raise MissingDataEntity(  # pragma: no cover
                ('Could not find a TestSample with {} == {} for '
                 'BioSample {}'.format(
                     KEY_EXTRACTION_TYPE, ext_type, self.bio_sample)))

    def __repr__(self):
        return 'CancerBioSample({})'.format(', '.join(
            map(str, [self.bio_entity, self.bio_sample])))

    def __str__(self):
        return repr(self)


class CancerDonor(GenericBioEntity):
    """Represent a donor in a matched tumor/normal"""

    # Override to use cancer-specific bio sample class
    bio_sample_class = CancerBioSample

    def __init__(self, shortcut_sheet, bio_entity):
        super().__init__(shortcut_sheet, bio_entity)
        #: The primary ``CancerMatchedSamplePair``
        self.primary_pair = self._get_primary_pair()
        #: All tumor/normal pairs
        self.all_pairs = list(self._iter_all_pairs())

    def _get_primary_pair(self):
        """Return primary ``CancerMatchedSamplePair``"""
        try:
            return next(self._iter_all_pairs())
        except StopIteration:
            return None

    def _iter_all_pairs(self):
        """Iterate all tumor/normal pair"""
        normal_bio_sample = self._get_primary_normal_bio_sample()
        if normal_bio_sample:
            for tumor_bio_sample in self._iter_tumor_bio_samples():
                yield CancerMatchedSamplePair(
                    self, tumor_bio_sample, normal_bio_sample)

    def _get_primary_normal_bio_sample(self):
        """Return primary normal ``BioSample``

        Raises ``MissingDataEntity`` in the case of problems
        """
        for bio_sample in self.bio_samples.values():
            if KEY_IS_TUMOR not in bio_sample.extra_infos:
                raise MissingDataEntity(  # pragma: no cover
                    'Could not find "{}" flag in BioSample {}'.format(
                        KEY_IS_TUMOR, bio_sample))
            elif not bio_sample.extra_infos[KEY_IS_TUMOR]:
                return bio_sample
        # Having no normal sample is an error by default but this behaviour
        # can be switched off.
        tpl = 'Could not find primary normal sample for BioEntity {}'
        msg = tpl.format(self.bio_entity)
        if not self.sheet.options.allow_missing_normal:  # pragma: no cover
            raise MissingDataEntity(msg)
        else:
            warn(msg, MissingDataWarning)

    def _iter_tumor_bio_samples(self):
        """Return iterable over all cancer bio samples

        The order depends on the order in ``self.bio_samples``.  If
        the type of this attribute is an ordered dict, then the behaviour of
        this function is reproducible, otherwise it is not.

        Raises ``MissingDataEntity`` in the case of problems
        """
        yielded_any = False
        for bio_sample in self.bio_samples.values():
            if KEY_IS_TUMOR not in bio_sample.extra_infos:
                raise MissingDataEntity(  # pragma: no cover
                    'Could not find "{}" flag in BioSample {}'.format(
                        KEY_IS_TUMOR, bio_sample))
            elif bio_sample.extra_infos[KEY_IS_TUMOR]:
                yielded_any = True
                yield bio_sample
        if not yielded_any:
            # Having no tumor sample is an error by default but this behaviour
            # can be switched off.
            tpl = 'Could not find a BioSample with {} = true for BioEntity {}'
            msg = tpl.format(KEY_IS_TUMOR, self.bio_entity)
            if not self.sheet.options.allow_missing_tumor:
                raise MissingDataEntity(msg)  # pragma: no cover
            else:
                warn(msg, MissingDataWarning)

    def __repr__(self):
        return 'CancerDonor({})'.format(', '.join(map(
            str, [self.sheet, self.bio_entity])))

    def __str__(self):
        return repr(self)



# -*- coding: utf-8 -*-
"""Tests for the shortcuts module with cancer sample sheet"""

import collections
import os
import pytest

from biomedsheets import io, ref_resolver, shortcuts


__author__ = 'Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>'


@pytest.fixture
def sheet_cancer():
    """Return ``Sheet`` instance for the cancer example"""
    path = os.path.join(os.path.abspath(
        os.path.dirname(__file__)), 'data', 'example_cancer.json')
    sheet_json = io.json_loads_ordered(open(path, 'rt').read())
    resolver = ref_resolver.RefResolver(dict_class=collections.OrderedDict)
    return io.SheetBuilder(
        resolver.resolve('file://' + path, sheet_json)).run()


def test_sheet_donors(sheet_cancer):
    """Tests for the sheet's ``donors`` attribute"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    assert len(cancer_cases.donors) == 2
    assert str(cancer_cases.donors[0].bio_entity.pk) == '1'
    assert str(cancer_cases.donors[0].bio_entity.secondary_id) == 'EX_001'
    assert str(cancer_cases.donors[1].bio_entity.pk) == '8'
    assert str(cancer_cases.donors[1].bio_entity.secondary_id) == 'EX_002'


def test_donor_shortcuts(sheet_cancer):
    """Tests for shortcuts of the Donor objects"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    # First donor
    assert str(cancer_cases.donors[0].pk) == '1'
    assert str(cancer_cases.donors[0].secondary_id) == 'EX_001'
    assert str(cancer_cases.donors[0].name) == 'EX_001-000001'
    assert cancer_cases.donors[0].primary_pair
    assert len(cancer_cases.donors[0].all_pairs) == 1
    assert cancer_cases.donors[0].enabled
    assert not cancer_cases.donors[0].disabled
    # Second donor
    assert str(cancer_cases.donors[1].pk) == '8'
    assert str(cancer_cases.donors[1].secondary_id) == 'EX_002'
    assert str(cancer_cases.donors[1].name) == 'EX_002-000008'
    assert cancer_cases.donors[1].primary_pair
    assert len(cancer_cases.donors[1].all_pairs) == 1
    assert cancer_cases.donors[1].enabled
    assert not cancer_cases.donors[1].disabled


def test_sample_pair_shortcuts(sheet_cancer):
    """Test shortcuts of the ``CancerMatchedSamplePair`` objects

    Also tests the shortcuts of ``CancerSample``
    """
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    pair = cancer_cases.donors[0].primary_pair
    assert str(pair).startswith('CancerMatchedSamplePair(')
    assert pair.donor.name == 'EX_001-000001'
    # Tumor sample: test samples
    assert pair.tumor_sample.name == 'EX_001-T1-000005'
    assert pair.tumor_sample.dna_test_sample.name == 'EX_001-T1-DNA1-000006'
    assert pair.tumor_sample.dna_test_sample.assay_sample.name == 'EX_001-T1-DNA1-WES1-000007'
    assert pair.tumor_sample.rna_test_sample is None
    # Normal sample: test samples
    assert pair.normal_sample.name == 'EX_001-N1-000002'
    assert pair.normal_sample.dna_test_sample.name == 'EX_001-N1-DNA1-000003'
    assert pair.normal_sample.dna_test_sample.assay_sample.name == 'EX_001-N1-DNA1-WES1-000004'
    assert pair.normal_sample.rna_test_sample is None


def test_sample_pairs_all(sheet_cancer):
    """Test for the sheet's ``all_sample_pairs`` attribute"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    assert len(cancer_cases.all_sample_pairs) == 2
    assert cancer_cases.all_sample_pairs[0].donor.name == 'EX_001-000001'
    assert cancer_cases.all_sample_pairs[0].tumor_sample.name == 'EX_001-T1-000005'
    assert cancer_cases.all_sample_pairs[0].normal_sample.name == 'EX_001-N1-000002'
    assert cancer_cases.all_sample_pairs[1].tumor_sample.name == 'EX_002-T1-000012'
    assert cancer_cases.all_sample_pairs[1].normal_sample.name == 'EX_002-N1-000009'


def test_sample_pairs_primary(sheet_cancer):
    """Test for the sheet's ``primary_sample_pairs`` attribute"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    assert len(cancer_cases.primary_sample_pairs) == 2
    assert cancer_cases.primary_sample_pairs[0].donor.name == 'EX_001-000001'
    assert cancer_cases.primary_sample_pairs[0].tumor_sample.name == 'EX_001-T1-000005'
    assert cancer_cases.primary_sample_pairs[0].normal_sample.name == 'EX_001-N1-000002'
    assert cancer_cases.primary_sample_pairs[1].tumor_sample.name == 'EX_002-T1-000012'
    assert cancer_cases.primary_sample_pairs[1].normal_sample.name == 'EX_002-N1-000009'


def test_cancer_donor(sheet_cancer):
    """Test for ``CancerDonor``"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    donor = cancer_cases.donors[0]
    assert str(donor).startswith('CancerDonor(')
    assert donor.primary_pair
    assert donor.primary_pair.tumor_sample.name == 'EX_001-T1-000005'
    assert donor.primary_pair.normal_sample.name == 'EX_001-N1-000002'
    assert len(donor.all_pairs) == 1


def test_cancer_bio_sample(sheet_cancer):
    """Test for ``CancerBioSample``"""
    cancer_cases = shortcuts.CancerCaseSheet(sheet_cancer)
    tumor_sample = cancer_cases.donors[0].primary_pair.tumor_sample
    assert tumor_sample.name == 'EX_001-T1-000005'
    assert tumor_sample.is_tumor
    normal_sample = cancer_cases.donors[0].primary_pair.normal_sample
    assert normal_sample.name == 'EX_001-N1-000002'
    assert not normal_sample.is_tumor
